import random
import time
from datetime import timedelta

from django.contrib.auth import get_user_model
from django.core.management.base import BaseCommand
from django.utils import timezone
from faker import Factory

from ....categories.models import Category
from ....core.pgutils import chunk_queryset
from ....threads.checksums import update_post_checksum
from ....threads.models import Thread
from ....users.models import Rank
from ...posts import get_fake_hidden_post, get_fake_post, get_fake_unapproved_post
from ...threads import (
    get_fake_closed_thread,
    get_fake_hidden_thread,
    get_fake_thread,
    get_fake_unapproved_thread,
)
from ...users import (
    get_fake_admin_activated_user,
    get_fake_banned_user,
    get_fake_deleted_user,
    get_fake_inactive_user,
    get_fake_user,
)

User = get_user_model()


class Command(BaseCommand):
    help = "Creates fake forum history reaching specified period."

    def add_arguments(self, parser):
        parser.add_argument(
            "length",
            help="generated history length (in days)",
            nargs="?",
            type=int,
            default=5,
        )
        parser.add_argument(
            "max_actions",
            help="number of items generate for a single day",
            nargs="?",
            type=int,
            default=50,
        )

    def handle(self, *args, **options):  # pylint: disable=too-many-locals
        history_length = options["length"]
        max_actions = options["max_actions"]
        fake = Factory.create()

        categories = list(Category.objects.all_categories())
        ranks = list(Rank.objects.all())

        message = "Creating fake forum history for %s days...\n"
        self.stdout.write(message % history_length)

        start_time = time.time()

        self.move_existing_users_to_past(history_length)

        start_timestamp = timezone.now()
        for days_ago in reversed(range(history_length)):
            date = start_timestamp - timedelta(days=days_ago)
            for date_variation in get_random_date_variations(date, 0, max_actions):
                action = random.randint(0, 100)
                if action >= 80:
                    self.create_fake_user(fake, date_variation, ranks)
                elif action > 50:
                    self.create_fake_thread(fake, date_variation, categories)
                else:
                    self.create_fake_post(fake, date_variation)

                if random.randint(0, 100) > 80:
                    self.create_fake_follow(date)

        self.synchronize_threads()
        self.synchronize_categories()

        total_time = time.time() - start_time
        total_humanized = time.strftime("%H:%M:%S", time.gmtime(total_time))
        message = "\n\nSuccessfully created fake history for %s days in %s"
        self.stdout.write(message % (history_length, total_humanized))

    def move_existing_users_to_past(self, history_length):
        for user in User.objects.all():
            user.joined_on -= timedelta(days=history_length)
            user.save(update_fields=["joined_on"])
            user.audittrail_set.all().delete()

    def create_fake_user(self, fake, date, ranks):
        # Pick random rank for next user
        rank = random.choice(ranks)

        # There's 10% chance user is inactive
        if random.randint(0, 100) > 90:
            user = get_fake_inactive_user(fake, rank)

        # There's another 10% chance user is admin-activated
        elif random.randint(0, 100) > 90:
            user = get_fake_admin_activated_user(fake, rank)

        # And further chance user is banned
        elif random.randint(0, 100) > 90:
            user = get_fake_banned_user(fake, rank)

        # Or deleted their account
        elif random.randint(0, 100) > 90:
            user = get_fake_deleted_user(fake, rank)

        # User is active
        else:
            user = get_fake_user(fake, rank)

        user.joined_on = date
        user.save(update_fields=["joined_on"])
        user.audittrail_set.all().delete()

        self.write_event(date, "%s has joined" % user)

    def create_fake_thread(self, fake, date, categories):
        category = random.choice(categories)

        # 10% chance thread poster is anonymous
        if random.randint(0, 100) > 90:
            starter = None
        else:
            starter = self.get_random_user(date)

        # There's 10% chance thread is closed
        if random.randint(0, 100) > 90:
            thread = get_fake_closed_thread(fake, category, starter)

        # There's further 5% chance thread is hidden
        elif random.randint(0, 100) > 95:
            if random.randint(0, 100) > 90:
                hidden_by = None
            else:
                hidden_by = self.get_random_user(date)

            thread = get_fake_hidden_thread(fake, category, starter, hidden_by)

        # And further 5% chance thread is unapproved
        elif random.randint(0, 100) > 95:
            thread = get_fake_unapproved_thread(fake, category, starter)

        # Default, standard thread
        else:
            thread = get_fake_thread(fake, category, starter)

        thread.first_post.posted_on = date
        thread.first_post.updated_on = date
        thread.first_post.checksum = update_post_checksum(thread.first_post)
        thread.first_post.save(update_fields=["checksum", "posted_on", "updated_on"])

        thread.started_on = date
        thread.save(update_fields=["started_on"])

        self.write_event(
            date, '%s has started "%s" thread' % (thread.first_post.poster_name, thread)
        )

    def create_fake_post(self, fake, date):
        thread = self.get_random_thread(date)
        if not thread:
            return

        # 10% chance poster is anonymous
        if random.randint(0, 100) > 90:
            poster = None
        else:
            poster = self.get_random_user(date)

        # There's 5% chance post is unapproved
        if random.randint(0, 100) > 90:
            post = get_fake_unapproved_post(fake, thread, poster)

        # There's further 5% chance post is hidden
        elif random.randint(0, 100) > 95:
            if random.randint(0, 100) > 90:
                hidden_by = None
            else:
                hidden_by = self.get_random_user(date)

            post = get_fake_hidden_post(fake, thread, poster, hidden_by)

        # Default, standard post
        else:
            post = get_fake_post(fake, thread, poster)

        post.posted_on = date
        post.updated_on = date
        post.checksum = update_post_checksum(post)
        post.save(update_fields=["checksum", "posted_on", "updated_on"])

        self.write_event(
            date, '%s has replied to "%s" thread' % (post.poster_name, thread)
        )

    def create_fake_follow(self, date):
        user_a = self.get_random_user(date)
        user_b = self.get_random_user(date)

        if not (user_a or user_b) or user_a == user_b:
            return

        if not user_a.is_following(user_b):
            user_a.follows.add(user_b)

        self.write_event(date, "%s followed %s" % (user_a, user_b))

    def get_random_thread(self, date):
        return (
            Thread.objects.filter(started_on__lt=date)
            .select_related("category")
            .order_by("?")
            .first()
        )

    def get_random_user(self, date):
        return (
            User.objects.filter(
                joined_on__lt=date, requires_activation=User.ACTIVATION_NONE
            )
            .order_by("?")
            .first()
        )

    def write_event(self, date, event):
        formatted_date = date.strftime("%Y-%m-%d %H:%M")
        self.stdout.write("%s: %s" % (formatted_date, event))

    def synchronize_threads(self):
        self.stdout.write("\nSynchronizing threads...")
        start_time = time.time()

        for thread in chunk_queryset(Thread.objects.all()):
            thread.synchronize()
            thread.save()

        total_time = time.time() - start_time
        total_humanized = time.strftime("%H:%M:%S", time.gmtime(total_time))

        message = "Synchronized %s threads in %s"
        self.stdout.write(message % (Thread.objects.count(), total_humanized))

    def synchronize_categories(self):
        self.stdout.write("\nSynchronizing categories...")
        start_time = time.time()

        for category in Category.objects.all_categories():
            category.synchronize()
            category.save()

        total_time = time.time() - start_time
        total_humanized = time.strftime("%H:%M:%S", time.gmtime(total_time))

        message = "Synchronized %s categories in %s"
        self.stdout.write(message % (Category.objects.count(), total_humanized))


def get_random_date_variations(date, min_date, max_date):
    variations = []
    for _ in range(random.randint(min_date, max_date)):
        random_offset = timedelta(minutes=random.randint(1, 1200))
        variations.append(date - random_offset)
    return sorted(variations)



# -*- coding: utf-8 -*-
from django.db import migrations

from .. import ACL_CACHE
from ...cache.operations import StartCacheVersioning


class Migration(migrations.Migration):

    dependencies = [
        ("misago_acl", "0003_default_roles"),
        ("misago_cache", "0001_initial"),
    ]

    operations = [StartCacheVersioning(ACL_CACHE)]



from .activetheme import get_active_theme
from .cache import get_theme_cache, set_theme_cache


def theme(request):
    active_theme = get_theme_cache(request.cache_versions)
    if active_theme is None:
        active_theme = get_active_theme()
        set_theme_cache(request.cache_versions, active_theme)

    return {"theme": active_theme}



from django.utils.translation import gettext_lazy
from rest_framework import serializers

from . import PostingEndpoint, PostingMiddleware
from ....markup import common_flavour
from ....users.audittrail import create_audit_trail
from ...checksums import update_post_checksum
from ...validators import validate_post, validate_post_length, validate_thread_title


class ReplyMiddleware(PostingMiddleware):
    def get_serializer(self):
        if self.mode == PostingEndpoint.START:
            serializer = ThreadSerializer
        else:
            serializer = ReplySerializer

        return serializer(data=self.request.data, context=self.kwargs)

    def save(self, serializer):
        if self.mode == PostingEndpoint.START:
            self.new_thread(serializer.validated_data)

        parsing_result = serializer.validated_data["parsing_result"]

        if self.mode == PostingEndpoint.EDIT:
            self.edit_post(serializer.validated_data, parsing_result)
        else:
            self.new_post(serializer.validated_data, parsing_result)

        if self.mode == PostingEndpoint.START:
            self.post.set_search_document(self.thread.title)
        else:
            self.post.set_search_document()

        self.post.updated_on = self.datetime
        self.post.save()

        self.post.update_search_vector()
        update_post_checksum(self.post)

        self.post.update_fields += ["checksum", "search_vector"]

        if self.mode == PostingEndpoint.START:
            self.thread.set_first_post(self.post)
            self.thread.set_last_post(self.post)

        self.thread.save()

        create_audit_trail(self.request, self.post)

        # annotate post for future middlewares
        self.post.parsing_result = parsing_result

    def new_thread(self, validated_data):
        self.thread.set_title(validated_data["title"])
        self.thread.starter_name = self.user.username
        self.thread.starter_slug = self.user.slug
        self.thread.last_poster_name = self.user.username
        self.thread.last_poster_slug = self.user.slug
        self.thread.started_on = self.datetime
        self.thread.last_post_on = self.datetime
        self.thread.save()

    def edit_post(self, validated_data, parsing_result):
        self.post.original = parsing_result["original_text"]
        self.post.parsed = parsing_result["parsed_text"]

    def new_post(self, validated_data, parsing_result):
        self.post.thread = self.thread
        self.post.poster = self.user
        self.post.poster_name = self.user.username
        self.post.posted_on = self.datetime

        self.post.original = parsing_result["original_text"]
        self.post.parsed = parsing_result["parsed_text"]


class ReplySerializer(serializers.Serializer):
    post = serializers.CharField(
        error_messages={"required": gettext_lazy("You have to enter a message.")}
    )

    def validate_post(self, data):
        validate_post_length(self.context["settings"], data)
        return data

    def validate(self, data):
        if data.get("post"):
            data["parsing_result"] = self.parse_post(data["post"])
            data = validate_post(self.context, data)

        return data

    def parse_post(self, post):
        if self.context["mode"] == PostingEndpoint.START:
            return common_flavour(self.context["request"], self.context["user"], post)

        return common_flavour(
            self.context["request"], self.context["post"].poster, post
        )


class ThreadSerializer(ReplySerializer):
    title = serializers.CharField(
        error_messages={"required": gettext_lazy("You have to enter thread title.")}
    )

    def validate_title(self, data):
        validate_thread_title(self.context["settings"], data)
        return data



# Generated by Django 1.11.15 on 2018-08-16 14:22
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("misago_legal", "0002_agreement_useragreement")]

    operations = []



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots[
    "test_multi_line_code_markdown 1"
] = """<pre><code>&lt;script&gt;
alert(&quot;!&quot;)
&lt;/script&gt;
</code></pre>"""

snapshots[
    "test_multi_line_code_markdown_with_language 1"
] = """<pre><code class="language-javascript">&lt;script&gt;
alert(&quot;!&quot;)
&lt;/script&gt;
</code></pre>"""

snapshots[
    "test_single_line_code_markdown 1"
] = "<p><code>&lt;script&gt;alert(&quot;!&quot;)&lt;/script&gt;</code></p>"



from datetime import timedelta
from io import StringIO

from django.contrib.auth import get_user_model
from django.core.management import call_command
from django.test import TestCase
from django.utils import timezone

from ...conf.test import override_dynamic_settings
from ..management.commands import deleteinactiveusers
from ..models import DeletedUser
from ..test import create_test_user

User = get_user_model()


class DeleteInactiveUsersTests(TestCase):
    def setUp(self):
        self.user = create_test_user("User", "user@example.com")

    @override_dynamic_settings(new_inactive_accounts_delete=2)
    def test_delete_user_activation_user(self):
        """deletes user that didn't activate their account within required time"""
        self.user.joined_on = timezone.now() - timedelta(days=2)
        self.user.requires_activation = User.ACTIVATION_USER
        self.user.save()

        out = StringIO()
        call_command(deleteinactiveusers.Command(), stdout=out)
        command_output = out.getvalue().splitlines()[0].strip()

        self.assertEqual(command_output, "Deleted inactive user accounts: 1")

        with self.assertRaises(User.DoesNotExist):
            self.user.refresh_from_db()

    @override_dynamic_settings(new_inactive_accounts_delete=2)
    def test_delete_user_activation_admin(self):
        """deletes user that wasn't activated by admin within required time"""
        self.user.joined_on = timezone.now() - timedelta(days=2)
        self.user.requires_activation = User.ACTIVATION_ADMIN
        self.user.save()

        out = StringIO()
        call_command(deleteinactiveusers.Command(), stdout=out)
        command_output = out.getvalue().splitlines()[0].strip()

        self.assertEqual(command_output, "Deleted inactive user accounts: 1")

        with self.assertRaises(User.DoesNotExist):
            self.user.refresh_from_db()

    @override_dynamic_settings(new_inactive_accounts_delete=2)
    def test_inactive_user_deletion_is_recorded(self):
        self.user.joined_on = timezone.now() - timedelta(days=2)
        self.user.requires_activation = User.ACTIVATION_USER
        self.user.save()

        out = StringIO()
        call_command(deleteinactiveusers.Command(), stdout=out)
        command_output = out.getvalue().splitlines()[0].strip()

        self.assertEqual(command_output, "Deleted inactive user accounts: 1")

        DeletedUser.objects.get(deleted_by=DeletedUser.DELETED_BY_SYSTEM)

    @override_dynamic_settings(new_inactive_accounts_delete=2)
    def test_skip_new_user_activation_user(self):
        """skips inactive user that is too new"""
        self.user.joined_on = timezone.now() - timedelta(days=1)
        self.user.requires_activation = User.ACTIVATION_USER
        self.user.save()

        out = StringIO()
        call_command(deleteinactiveusers.Command(), stdout=out)
        command_output = out.getvalue().splitlines()[0].strip()

        self.assertEqual(command_output, "Deleted inactive user accounts: 0")

        self.user.refresh_from_db()

    @override_dynamic_settings(new_inactive_accounts_delete=2)
    def test_skip_new_user_activation_admin(self):
        """skips admin-activated user that is too new"""
        self.user.joined_on = timezone.now() - timedelta(days=1)
        self.user.requires_activation = User.ACTIVATION_ADMIN
        self.user.save()

        out = StringIO()
        call_command(deleteinactiveusers.Command(), stdout=out)
        command_output = out.getvalue().splitlines()[0].strip()

        self.assertEqual(command_output, "Deleted inactive user accounts: 0")

        self.user.refresh_from_db()

    @override_dynamic_settings(new_inactive_accounts_delete=2)
    def test_skip_active_user(self):
        """skips active user"""
        self.user.joined_on = timezone.now() - timedelta(days=1)
        self.user.save()

        out = StringIO()
        call_command(deleteinactiveusers.Command(), stdout=out)
        command_output = out.getvalue().splitlines()[0].strip()

        self.assertEqual(command_output, "Deleted inactive user accounts: 0")

        self.user.refresh_from_db()

    @override_dynamic_settings(new_inactive_accounts_delete=0)
    def test_delete_inactive_is_disabled(self):
        """skips active user"""
        self.user.joined_on = timezone.now() - timedelta(days=1)
        self.user.requires_activation = User.ACTIVATION_ADMIN
        self.user.save()

        out = StringIO()
        call_command(deleteinactiveusers.Command(), stdout=out)
        command_output = out.getvalue().splitlines()[0].strip()

        self.assertEqual(
            command_output,
            "Automatic deletion of inactive user accounts is currently disabled.",
        )

        self.user.refresh_from_db()



from django.urls import path

from . import views


urlpatterns = [
    path("login/<str:backend>/", views.auth, name="social-begin"),
    path("complete/<str:backend>/", views.complete, name="social-complete"),
]



import os

import pytest
from django.core.files.uploadedfile import SimpleUploadedFile

from ...models import Icon


@pytest.fixture
def favicon(db, image_alt):
    return Icon.objects.create(
        type=Icon.TYPE_FAVICON,
        image=SimpleUploadedFile("favicon.png", image_alt, "image/png"),
    )


@pytest.fixture
def favicon_32(db, image_alt):
    return Icon.objects.create(
        type=Icon.TYPE_FAVICON_32,
        image=SimpleUploadedFile("favicon-32.png", image_alt, "image/png"),
    )


@pytest.fixture
def favicon_16(db, image_alt):
    return Icon.objects.create(
        type=Icon.TYPE_FAVICON_16,
        image=SimpleUploadedFile("favicon-16.png", image_alt, "image/png"),
    )


def test_uploading_favicon_sets_favicon_images(admin_client, admin_link, image):
    admin_client.post(
        admin_link, {"favicon": SimpleUploadedFile("image.png", image, "image/png")}
    )
    Icon.objects.get(type=Icon.TYPE_FAVICON)
    Icon.objects.get(type=Icon.TYPE_FAVICON_32)
    Icon.objects.get(type=Icon.TYPE_FAVICON_16)


def test_uploading_favicon_removes_existing_favicon_images(
    admin_client, admin_link, image, favicon, favicon_32, favicon_16
):
    admin_client.post(
        admin_link, {"favicon": SimpleUploadedFile("image.png", image, "image/png")}
    )

    with pytest.raises(Icon.DoesNotExist):
        favicon.refresh_from_db()

    with pytest.raises(Icon.DoesNotExist):
        favicon_32.refresh_from_db()

    with pytest.raises(Icon.DoesNotExist):
        favicon_16.refresh_from_db()


def test_uploading_new_favicon_removes_old_one_image_file(
    admin_client, admin_link, image, favicon, favicon_32, favicon_16
):
    admin_client.post(
        admin_link, {"favicon": SimpleUploadedFile("image.png", image, "image/png")}
    )

    assert not os.path.exists(favicon.image.path)
    assert not os.path.exists(favicon_32.image.path)
    assert not os.path.exists(favicon_16.image.path)


def test_submitting_form_without_new_icon_does_not_delete_old_favicon_images(
    admin_client, admin_link, favicon, favicon_32, favicon_16
):
    admin_client.post(admin_link, {})
    favicon.refresh_from_db()
    favicon_32.refresh_from_db()
    favicon_16.refresh_from_db()


def test_favicon_can_be_deleted_without_setting_new_one(
    admin_client, admin_link, favicon, favicon_32, favicon_16
):
    admin_client.post(admin_link, {"favicon_delete": "1"})

    with pytest.raises(Icon.DoesNotExist):
        favicon.refresh_from_db()

    with pytest.raises(Icon.DoesNotExist):
        favicon_32.refresh_from_db()

    with pytest.raises(Icon.DoesNotExist):
        favicon_16.refresh_from_db()


def test_deleting_icon_also_deletes_its_image_files(
    admin_client, admin_link, favicon, favicon_32, favicon_16
):
    admin_client.post(admin_link, {"favicon_delete": "1"})
    assert not os.path.exists(favicon.image.path)
    assert not os.path.exists(favicon_32.image.path)
    assert not os.path.exists(favicon_16.image.path)


def test_uploading_invalid_icon_does_not_remove_current_icon(
    admin_client, admin_link, favicon, favicon_32, favicon_16, image_small
):
    admin_client.post(
        admin_link,
        {"favicon": SimpleUploadedFile("image.png", image_small, "image/png")},
    )

    favicon.refresh_from_db()
    favicon_32.refresh_from_db()
    favicon_16.refresh_from_db()


def test_icon_is_not_set_because_it_was_not_square(
    admin_client, admin_link, image_non_square
):
    admin_client.post(
        admin_link,
        {"favicon": SimpleUploadedFile("image.png", image_non_square, "image/png")},
    )

    assert not Icon.objects.exists()


def test_icon_is_not_set_because_it_was_too_small(
    admin_client, admin_link, image_small
):
    admin_client.post(
        admin_link,
        {"favicon": SimpleUploadedFile("image.png", image_small, "image/png")},
    )

    assert not Icon.objects.exists()



import pytest

from ..models import Setting


@pytest.fixture
def lazy_setting(db):
    return Setting.objects.create(
        setting="lazy_setting", dry_value="Hello", is_lazy=True
    )


@pytest.fixture
def lazy_setting_without_value(db):
    return Setting.objects.create(setting="lazy_setting", dry_value="", is_lazy=True)


@pytest.fixture
def private_setting(db):
    return Setting.objects.create(
        setting="private_setting", dry_value="Hello", is_public=False
    )


@pytest.fixture
def public_setting(db):
    return Setting.objects.create(
        setting="public_setting", dry_value="Hello", is_public=True
    )



from django.http import Http404
from django.test.client import RequestFactory
from django.urls import reverse

from ...acl.useracl import get_user_acl
from ...conf.dynamicsettings import DynamicSettings
from ...conftest import get_cache_versions
from ...users.models import AnonymousUser
from ..middleware import ExceptionHandlerMiddleware


def create_request():
    request = RequestFactory().get(reverse("misago:index"))
    request.cache_versions = get_cache_versions()
    request.settings = DynamicSettings(request.cache_versions)
    request.user = AnonymousUser()
    request.user_acl = get_user_acl(request.user, request.cache_versions)
    request.include_frontend_context = True
    request.frontend_context = {}
    request.socialauth = {}
    return request


def get_response(*_args):
    raise NotImplementedError()


def test_exception_handler_middleware_returns_response_for_supported_exception(db):
    """Middleware returns HttpResponse for supported exception"""
    middleware = ExceptionHandlerMiddleware(get_response)
    exception = Http404()
    assert middleware.process_exception(create_request(), exception)


def test_exception_handler_middleware_returns_none_for_non_supported_exception(db):
    """Middleware returns None for non-supported exception"""
    middleware = ExceptionHandlerMiddleware(get_response)
    exception = TypeError()
    assert middleware.process_exception(create_request(), exception) is None



from django.contrib.auth import get_user_model
from django.urls import reverse

from ...admin.test import AdminTestCase

User = get_user_model()


class GenderProfileFieldTests(AdminTestCase):
    def setUp(self):
        super().setUp()

        self.test_link = reverse("misago:admin:users:edit", kwargs={"pk": self.user.pk})

    def test_field_displays_in_admin(self):
        """field displays in admin"""
        response = self.client.get(self.test_link)
        self.assertContains(response, 'name="gender"')

    def test_admin_clears_field(self):
        """admin form allows admins to clear field"""
        self.user.profile_fields["gender"] = "female"
        self.user.save()

        self.reload_user()
        self.assertEqual(self.user.profile_fields["gender"], "female")

        response = self.client.post(
            self.test_link,
            data={
                "username": "Edited",
                "rank": str(self.user.rank_id),
                "roles": str(self.user.roles.all()[0].pk),
                "email": "reg@stered.com",
                "new_password": "",
                "signature": "",
                "is_signature_locked": "0",
                "is_hiding_presence": "0",
                "limits_private_thread_invites_to": "0",
                "signature_lock_staff_message": "",
                "signature_lock_user_message": "",
                "subscribe_to_started_threads": "2",
                "subscribe_to_replied_threads": "2",
            },
        )
        self.assertEqual(response.status_code, 302)

        self.reload_user()
        self.assertEqual(self.user.profile_fields["gender"], "")

    def test_admin_validates_field(self):
        """admin form allows admins to edit field"""
        response = self.client.post(
            self.test_link,
            data={
                "username": "Edited",
                "rank": str(self.user.rank_id),
                "roles": str(self.user.roles.all()[0].pk),
                "email": "reg@stered.com",
                "gender": "attackcopter",
                "new_password": "",
                "signature": "",
                "is_signature_locked": "0",
                "is_hiding_presence": "0",
                "limits_private_thread_invites_to": "0",
                "signature_lock_staff_message": "",
                "signature_lock_user_message": "",
                "subscribe_to_started_threads": "2",
                "subscribe_to_replied_threads": "2",
            },
        )

        self.assertContains(
            response, "attackcopter is not one of the available choices."
        )

    def test_admin_edits_field(self):
        """admin form allows admins to edit field"""
        response = self.client.post(
            self.test_link,
            data={
                "username": "Edited",
                "rank": str(self.user.rank_id),
                "roles": str(self.user.roles.all()[0].pk),
                "email": "reg@stered.com",
                "gender": "female",
                "new_password": "",
                "signature": "",
                "is_signature_locked": "0",
                "is_hiding_presence": "0",
                "limits_private_thread_invites_to": "0",
                "signature_lock_staff_message": "",
                "signature_lock_user_message": "",
                "subscribe_to_started_threads": "2",
                "subscribe_to_replied_threads": "2",
            },
        )
        self.assertEqual(response.status_code, 302)

        self.reload_user()
        self.assertEqual(self.user.profile_fields["gender"], "female")

    def test_admin_search_field(self):
        """admin users search searches this field"""
        test_link = reverse("misago:admin:users:index")

        response = self.client.get("%s?redirected=1&profilefields=female" % test_link)
        self.assertContains(response, "No users matching criteria exist.")

        # search by value
        self.user.profile_fields["gender"] = "female"
        self.user.save()

        response = self.client.get("%s?redirected=1&profilefields=female" % test_link)
        self.assertNotContains(response, "No users matching criteria exist.")

        # search by choice name
        self.user.profile_fields["gender"] = "secret"
        self.user.save()

        response = self.client.get("%s?redirected=1&profilefields=telling" % test_link)
        self.assertNotContains(response, "No users matching criteria exist.")

    def test_field_display(self):
        """field displays on user profile when filled in"""
        test_link = reverse(
            "misago:user-details", kwargs={"pk": self.user.pk, "slug": self.user.slug}
        )

        response = self.client.get(test_link)
        self.assertNotContains(response, "Gender")

        self.user.profile_fields["gender"] = "secret"
        self.user.save()

        response = self.client.get(test_link)
        self.assertContains(response, "Gender")
        self.assertContains(response, "Not telling")

    def test_field_outdated_hidden(self):
        """field with outdated value is hidden"""
        test_link = reverse(
            "misago:user-details", kwargs={"pk": self.user.pk, "slug": self.user.slug}
        )

        response = self.client.get(test_link)
        self.assertNotContains(response, "Gender")

        self.user.profile_fields["gender"] = "not valid"
        self.user.save()

        response = self.client.get(test_link)
        self.assertNotContains(response, "Gender")

    def test_field_display_json(self):
        """field is included in display json"""
        test_link = reverse("misago:api:user-details", kwargs={"pk": self.user.pk})

        response = self.client.get(test_link)
        self.assertEqual(
            response.json()["groups"],
            [
                {
                    "name": "IP address",
                    "fields": [
                        {"fieldname": "join_ip", "name": "Join IP", "text": "127.0.0.1"}
                    ],
                }
            ],
        )

        self.user.profile_fields["gender"] = "male"
        self.user.save()

        response = self.client.get(test_link)
        self.assertEqual(
            response.json()["groups"],
            [
                {
                    "name": "Personal",
                    "fields": [
                        {"fieldname": "gender", "name": "Gender", "text": "Male"}
                    ],
                },
                {
                    "name": "IP address",
                    "fields": [
                        {"fieldname": "join_ip", "name": "Join IP", "text": "127.0.0.1"}
                    ],
                },
            ],
        )

    def test_field_outdated_hidden_json(self):
        """field with outdated value is removed in display json"""
        test_link = reverse("misago:api:user-details", kwargs={"pk": self.user.pk})

        response = self.client.get(test_link)
        self.assertEqual(
            response.json()["groups"],
            [
                {
                    "name": "IP address",
                    "fields": [
                        {"fieldname": "join_ip", "name": "Join IP", "text": "127.0.0.1"}
                    ],
                }
            ],
        )

        self.user.profile_fields["gender"] = "invalid"
        self.user.save()

        response = self.client.get(test_link)
        self.assertEqual(
            response.json()["groups"],
            [
                {
                    "name": "IP address",
                    "fields": [
                        {"fieldname": "join_ip", "name": "Join IP", "text": "127.0.0.1"}
                    ],
                }
            ],
        )

    def test_api_returns_field_json(self):
        """field json is returned from API"""
        test_link = reverse("misago:api:user-edit-details", kwargs={"pk": self.user.pk})

        response = self.client.get(test_link)

        found_field = None
        for group in response.json():
            for field in group["fields"]:
                if field["fieldname"] == "gender":
                    found_field = field

        self.assertEqual(
            found_field,
            {
                "fieldname": "gender",
                "label": "Gender",
                "help_text": None,
                "input": {
                    "type": "select",
                    "choices": [
                        {"label": "Not specified", "value": ""},
                        {"label": "Not telling", "value": "secret"},
                        {"label": "Female", "value": "female"},
                        {"label": "Male", "value": "male"},
                    ],
                },
                "initial": "",
            },
        )

    def test_api_clears_field(self):
        """field can be cleared via api"""
        test_link = reverse("misago:api:user-edit-details", kwargs={"pk": self.user.pk})

        self.user.profile_fields["gender"] = "secret"
        self.user.save()

        self.reload_user()
        self.assertEqual(self.user.profile_fields["gender"], "secret")

        response = self.client.post(test_link, data={})
        self.assertEqual(response.status_code, 200)

        self.reload_user()
        self.assertEqual(self.user.profile_fields["gender"], "")

    def test_api_validates_field(self):
        """field can be edited via api"""
        test_link = reverse("misago:api:user-edit-details", kwargs={"pk": self.user.pk})

        response = self.client.post(test_link, data={"gender": "attackhelicopter"})
        self.assertContains(
            response,
            "attackhelicopter is not one of the available choices.",
            status_code=400,
        )

    def test_api_edits_field(self):
        """field can be edited via api"""
        test_link = reverse("misago:api:user-edit-details", kwargs={"pk": self.user.pk})

        response = self.client.post(test_link, data={"gender": "female"})
        self.assertEqual(response.status_code, 200)

        self.reload_user()
        self.assertEqual(self.user.profile_fields["gender"], "female")



from ..parser import parse


def test_bold_bbcode(request_mock, user, snapshot):
    text = "Lorem [b]ipsum[/b]!"
    result = parse(text, request_mock, user)
    snapshot.assert_match(result["parsed_text"])


def test_italics_bbcode(request_mock, user, snapshot):
    text = "Lorem [i]ipsum[/i]!"
    result = parse(text, request_mock, user)
    snapshot.assert_match(result["parsed_text"])


def test_underline_bbcode(request_mock, user, snapshot):
    text = "Lorem [u]ipsum[/u]!"
    result = parse(text, request_mock, user)
    snapshot.assert_match(result["parsed_text"])


def test_simple_inline_bbcode_is_escaped(request_mock, user, snapshot):
    text = 'Lorem [b]ips <script language="application/javascript"> um[/b]!'
    result = parse(text, request_mock, user)
    snapshot.assert_match(result["parsed_text"])


def test_inline_bbcode_can_be_mixed(request_mock, user, snapshot):
    text = "Lorem [b][u]ipsum[/u][/b]!"
    result = parse(text, request_mock, user)
    snapshot.assert_match(result["parsed_text"])


def test_inline_bbcode_can_be_mixed_with_markdown(request_mock, user, snapshot):
    text = "Lorem [b]**ipsum**[/b]!"
    result = parse(text, request_mock, user)
    snapshot.assert_match(result["parsed_text"])


def test_image_bbcode(request_mock, user, snapshot):
    text = "Lorem [img]https://placekitten.com/g/1200/500[/img] ipsum"
    result = parse(text, request_mock, user)
    snapshot.assert_match(result["parsed_text"])


def test_image_bbcode_is_escaped(request_mock, user, snapshot):
    text = 'Lorem [img]<script language="application/javascript">[/img] ipsum'
    result = parse(text, request_mock, user)
    snapshot.assert_match(result["parsed_text"])


def test_image_bbcode_is_case_insensitive(request_mock, user, snapshot):
    text = "Lorem [iMg]https://placekitten.com/g/1200/500[/ImG] ipsum"
    result = parse(text, request_mock, user)
    snapshot.assert_match(result["parsed_text"])


def test_url_bbcode(request_mock, user, snapshot):
    text = "Lorem [url]https://placekitten.com/g/1200/500[/url] ipsum"
    result = parse(text, request_mock, user)
    snapshot.assert_match(result["parsed_text"])


def test_url_bbcode_is_escaped(request_mock, user, snapshot):
    text = 'Lorem [url]<script language="application/javascript">[/url] ipsum'
    result = parse(text, request_mock, user)
    snapshot.assert_match(result["parsed_text"])


def test_url_bbcode_with_link_text(request_mock, user, snapshot):
    text = "Lorem [url=https://placekitten.com/g/1200/500]dolor[/url] ipsum"
    result = parse(text, request_mock, user)
    snapshot.assert_match(result["parsed_text"])


def test_url_bbcode_with_long_link_text(request_mock, user, snapshot):
    text = "Lorem [url=https://placekitten.com/g/1200/500]dolor met[/url] ipsum"
    result = parse(text, request_mock, user)
    snapshot.assert_match(result["parsed_text"])


def test_url_bbcode_with_quotes_and_link_text(request_mock, user, snapshot):
    text = 'Lorem [url="https://placekitten.com/g/1200/500"]dolor[/url] ipsum'
    result = parse(text, request_mock, user)
    snapshot.assert_match(result["parsed_text"])


def test_url_bbcode_with_quotes_and_long_link_text(request_mock, user, snapshot):
    text = 'Lorem [url="https://placekitten.com/g/1200/500"]dolor met[/url] ipsum'
    result = parse(text, request_mock, user)
    snapshot.assert_match(result["parsed_text"])


def test_url_bbcode_link_text_is_escaped(request_mock, user, snapshot):
    text = (
        'Lorem [url=<script language="application/javascript">]'
        '<script language="application/javascript">'
        "[/url] ipsum"
    )

    result = parse(text, request_mock, user)
    snapshot.assert_match(result["parsed_text"])



from datetime import timedelta

from ariadne import QueryType
from django.contrib.auth import get_user_model
from django.core.cache import cache
from django.utils import timezone

from ...threads.models import Attachment, Post, Thread
from ...users.models import DataDownload, DeletedUser

CACHE_KEY = "misago_admin_analytics"
CACHE_LENGTH = 3600 * 4  # 4 hours

User = get_user_model()

analytics = QueryType()


@analytics.field("analytics")
def resolve_analytics(_, info, *, span):
    span = clean_span(span)
    cache_key = "%s_%s" % (CACHE_KEY, span)
    data = cache.get(cache_key)
    if not data:
        data = get_data_from_db(span)
        cache.set(cache_key, data, CACHE_LENGTH)
    return data


def clean_span(span):
    if span > 360:
        return 360
    if span < 30:
        return 30
    return span


def get_data_from_db(span):
    analytics = Analytics(span)

    return {
        "users": analytics.get_data_for_model(User, "joined_on"),
        "userDeletions": analytics.get_data_for_model(DeletedUser, "deleted_on"),
        "threads": analytics.get_data_for_model(Thread, "started_on"),
        "posts": analytics.get_data_for_model(Post, "posted_on"),
        "attachments": analytics.get_data_for_model(Attachment, "uploaded_on"),
        "dataDownloads": analytics.get_data_for_model(DataDownload, "requested_on"),
    }


class Analytics:
    def __init__(self, span):
        self.today = timezone.now()
        self.span = span

        self.cutoff = self.today - timedelta(days=span * 2)
        self.legend = self.get_legend()

    def get_legend(self):
        legend = []
        for day in range(self.span * 2):
            date = self.today - timedelta(days=day)
            legend.append(date.strftime("%x"))
        return legend

    def get_empty_data(self):
        return {k: 0 for k in self.legend}

    def get_data_for_model(self, model, date_attr):
        filter_kwarg = {"%s__gte" % date_attr: self.cutoff}
        queryset = model.objects.filter(**filter_kwarg).order_by("-pk")

        data = self.get_empty_data()
        for item in queryset.values(date_attr).iterator():
            date = item[date_attr].strftime("%x")
            if date in data:
                data[date] += 1

        values = list(data.values())
        current = list(reversed(values[: self.span]))
        previous = list(reversed(values[self.span :]))

        return {
            "current": current,
            "currentCumulative": cumulate_data(current),
            "previous": previous,
            "previousCumulative": cumulate_data(previous),
        }


def cumulate_data(data_series):
    data = []
    for v in data_series:
        if not data:
            data.append(v)
        else:
            data.append(data[-1] + v)
    return data



from django.urls import path
from django.utils.translation import gettext_lazy as _

from .views import (
    DeleteMenuItem,
    EditMenuItem,
    MenuItemsList,
    MoveDownMenuItem,
    MoveUpMenuItem,
    NewMenuItem,
)


class MisagoAdminExtension:
    def register_urlpatterns(self, urlpatterns):
        # Menu items
        urlpatterns.namespace("menu-items/", "menu-items", "settings")
        urlpatterns.patterns(
            "settings:menu-items",
            path("", MenuItemsList.as_view(), name="index"),
            path("<int:page>/", MenuItemsList.as_view(), name="index"),
            path("new/", NewMenuItem.as_view(), name="new"),
            path("edit/<int:pk>/", EditMenuItem.as_view(), name="edit"),
            path("delete/<int:pk>/", DeleteMenuItem.as_view(), name="delete"),
            path("down/<int:pk>/", MoveDownMenuItem.as_view(), name="down"),
            path("up/<int:pk>/", MoveUpMenuItem.as_view(), name="up"),
        )

    def register_navigation_nodes(self, site):
        site.add_node(
            name=_("Menu items"),
            description=_(
                "Use those options to add custom items to the navbar and footer menus."
            ),
            parent="settings",
            namespace="menu-items",
        )



import pytest
from django.urls import reverse

from ...models import SocialAuthProvider


admin_link = reverse("misago:admin:settings:socialauth:edit", kwargs={"pk": "github"})


@pytest.fixture
def provider(db):
    return SocialAuthProvider.objects.create(provider="github", is_active=True, order=0)


def test_github_form_can_be_accessed(admin_client):
    response = admin_client.get(admin_link)
    assert response.status_code == 200


def test_github_login_can_be_setup(admin_client):
    admin_client.post(
        admin_link,
        {
            "is_active": "1",
            "associate_by_email": "1",
            "key": "test-key",
            "secret": "test-secret",
        },
    )

    provider = SocialAuthProvider.objects.get(provider="github")
    assert provider.is_active
    assert provider.settings == {
        "associate_by_email": 1,
        "key": "test-key",
        "secret": "test-secret",
    }


def test_github_login_can_be_disabled(admin_client, provider):
    admin_client.post(admin_link, {"is_active": "0"})

    provider = SocialAuthProvider.objects.get(provider="github")
    assert not provider.is_active


def test_github_login_form_requires_key_to_setup(admin_client):
    admin_client.post(admin_link, {"is_active": "1", "secret": "test-secret"})

    with pytest.raises(SocialAuthProvider.DoesNotExist):
        SocialAuthProvider.objects.get(provider="github")


def test_github_login_form_requires_secret_to_setup(admin_client):
    admin_client.post(admin_link, {"is_active": "1", "key": "test-key"})

    with pytest.raises(SocialAuthProvider.DoesNotExist):
        SocialAuthProvider.objects.get(provider="github")



from django.core.paginator import EmptyPage, InvalidPage


def get_page(queryset, order_by, per_page, start=0):
    if start < 0:
        raise InvalidPage()

    object_list = list(_slice_queryset(queryset, order_by, per_page, start))
    if start and not object_list:
        raise EmptyPage()

    next_cursor = None
    if len(object_list) > per_page:
        next_slice_first_item = object_list.pop(-1)
        attr_name = order_by.lstrip("-")
        next_cursor = getattr(next_slice_first_item, attr_name)

    return CursorPage(start, object_list, next_cursor)


def _slice_queryset(queryset, order_by, per_page, start):
    page_len = int(per_page) + 1
    if start:
        if order_by.startswith("-"):
            filter_name = "%s__lte" % order_by[1:]
        else:
            filter_name = "%s__gte" % order_by
        return queryset.filter(**{filter_name: start})[:page_len]
    return queryset[:page_len]


class CursorPage:
    def __init__(self, start, object_list, next_=None):
        self.start = start or 0
        self.first = self.start == 0
        self.object_list = object_list
        self.next = next_

    def __len__(self):
        return len(self.object_list)

    def has_next(self):
        return bool(self.next)



from django.views import View

from .. import render


class AdminView(View):
    def get_template_name(self, request):
        return "%s/%s" % (self.templates_dir, self.template_name)

    def current_link(self, request):
        matched_url = request.resolver_match.url_name
        return "%s:%s" % (request.resolver_match.namespace, matched_url)

    def process_context(self, request, context):
        """simple hook for extending and manipulating template context."""
        return context

    def render(self, request, context=None, template_name=None):
        context = context or {}

        context["root_link"] = self.root_link
        context["current_link"] = self.current_link(request)

        context = self.process_context(request, context)

        template_name = template_name or self.get_template_name(request)
        return render(request, template_name, context)



# Generated by Django 2.2.3 on 2019-11-10 14:32

from django.db import migrations


def rename_google_provider(apps, _):
    SocialAuthProvider = apps.get_model("misago_socialauth", "SocialAuthProvider")
    SocialAuthProvider.objects.filter(provider="google").update(
        provider="google-oauth2"
    )


class Migration(migrations.Migration):

    dependencies = [("misago_socialauth", "0002_cache_version")]

    operations = [migrations.RunPython(rename_google_provider)]



from django.core import exceptions as django_exceptions
from django.core.exceptions import PermissionDenied
from django.http import Http404
from django.test import TestCase

from .. import exceptionhandler
from ...users.models import Ban
from ..exceptions import Banned

INVALID_EXCEPTIONS = [
    django_exceptions.ObjectDoesNotExist,
    django_exceptions.ViewDoesNotExist,
    TypeError,
    ValueError,
    KeyError,
]


class IsMisagoExceptionTests(TestCase):
    def test_is_misago_exception_true_for_handled_exceptions(self):
        """exceptionhandler.is_misago_exception recognizes handled exceptions"""
        for exception in exceptionhandler.HANDLED_EXCEPTIONS:
            self.assertTrue(exceptionhandler.is_misago_exception(exception()))

    def test_is_misago_exception_false_for_not_handled_exceptions(self):
        """exceptionhandler.is_misago_exception fails to recognize other exceptions"""
        for exception in INVALID_EXCEPTIONS:
            self.assertFalse(exceptionhandler.is_misago_exception(exception()))


class GetExceptionHandlerTests(TestCase):
    def test_exception_handlers_list(self):
        """HANDLED_EXCEPTIONS length matches that of EXCEPTION_HANDLERS"""
        self.assertEqual(
            len(exceptionhandler.HANDLED_EXCEPTIONS),
            len(exceptionhandler.EXCEPTION_HANDLERS),
        )

    def test_get_exception_handler_for_handled_exceptions(self):
        """Exception handler has correct handler for every Misago exception"""
        for exception in exceptionhandler.HANDLED_EXCEPTIONS:
            exceptionhandler.get_exception_handler(exception())

    def test_get_exception_handler_for_non_handled_exceptio(self):
        """Exception handler has no handler for non-supported exception"""
        for exception in INVALID_EXCEPTIONS:
            with self.assertRaises(ValueError):
                exceptionhandler.get_exception_handler(exception())


class HandleAPIExceptionTests(TestCase):
    def test_banned(self):
        """banned exception is correctly handled"""
        ban = Ban(user_message="This is test ban!")

        response = exceptionhandler.handle_api_exception(Banned(ban), None)

        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.data["ban"]["message"]["html"], "<p>This is test ban!</p>"
        )

    def test_permission_denied(self):
        """permission denied exception is correctly handled"""
        response = exceptionhandler.handle_api_exception(PermissionDenied(), None)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.data["detail"],
            "You do not have permission to perform this action.",
        )

    def test_permission_message_denied(self):
        """permission denied with message is correctly handled"""
        exception = PermissionDenied("You shall not pass!")
        response = exceptionhandler.handle_api_exception(exception, None)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(response.data["detail"], "You shall not pass!")

    def test_unhandled_exception(self):
        """our exception handler is not interrupting other exceptions"""
        for exception in INVALID_EXCEPTIONS:
            response = exceptionhandler.handle_api_exception(exception(), None)
            self.assertIsNone(response)

        response = exceptionhandler.handle_api_exception(Http404(), None)
        self.assertEqual(response.status_code, 404)
        self.assertEqual(response.data["detail"], "Not found.")



from django.urls import path
from django.utils.translation import gettext_lazy as _

from .views import DeleteRole, EditRole, NewRole, RolesList, RoleUsers


class MisagoAdminExtension:
    def register_urlpatterns(self, urlpatterns):
        # Permissions section
        urlpatterns.namespace("permissions/", "permissions")

        # Roles
        urlpatterns.patterns(
            "permissions",
            path("", RolesList.as_view(), name="index"),
            path("new/", NewRole.as_view(), name="new"),
            path("edit/<int:pk>/", EditRole.as_view(), name="edit"),
            path("users/<int:pk>/", RoleUsers.as_view(), name="users"),
            path("delete/<int:pk>/", DeleteRole.as_view(), name="delete"),
        )

    def register_navigation_nodes(self, site):
        site.add_node(
            name=_("Permissions"),
            icon="fa fa-adjust",
            after="categories:index",
            namespace="permissions",
        )



from django.urls import reverse

from ....test import assert_contains


def test_link_is_registered_in_admin_nav(admin_client):
    response = admin_client.get(reverse("misago:admin:index"))
    assert_contains(response, reverse("misago:admin:settings:index"))


def test_settings_grid_renders(admin_client):
    response = admin_client.get(reverse("misago:admin:settings:index"))
    assert response.status_code == 200



from ...conf.test import override_dynamic_settings
from ...test import assert_contains, assert_not_contains


@override_dynamic_settings(google_tracking_id=None)
def test_tracking_script_is_not_included_if_tracking_id_is_not_set(db, client):
    response = client.get("/")
    assert_not_contains(response, "googletagmanager.com/gtag/js")


@override_dynamic_settings(google_tracking_id="UA-TEST")
def test_tracking_script_is_included_if_tracking_id_is_not_set(db, client):
    response = client.get("/")
    assert_contains(response, "googletagmanager.com/gtag/js")
    assert_contains(response, "UA-TEST")



from time import time

from django.core.management.base import BaseCommand

from ...activepostersranking import build_active_posters_ranking


class Command(BaseCommand):
    help = "Builds active posters ranking"

    def handle(self, *args, **options):
        self.stdout.write("\nBuilding active posters ranking...")

        start_time = time()
        build_active_posters_ranking()
        end_time = time() - start_time

        self.stdout.write("Finished after %.2fs" % end_time)



import json

from django.core import mail

from .. import test
from ...acl.test import patch_user_acl
from ...users.test import create_test_user
from ..models import Thread, ThreadParticipant
from ..test import other_user_cant_use_private_threads
from .test_privatethreads import PrivateThreadsTestCase


class PrivateThreadPatchApiTestCase(PrivateThreadsTestCase):
    def setUp(self):
        super().setUp()

        self.thread = test.post_thread(self.category, poster=self.user)
        self.api_link = self.thread.get_api_url()

        self.other_user = create_test_user("OtherUser", "otheruser@example.com")

    def patch(self, api_link, ops):
        return self.client.patch(
            api_link, json.dumps(ops), content_type="application/json"
        )


class PrivateThreadAddParticipantApiTests(PrivateThreadPatchApiTestCase):
    def test_add_participant_not_owner(self):
        """non-owner can't add participant"""
        ThreadParticipant.objects.add_participants(self.thread, [self.user])

        response = self.patch(
            self.api_link,
            [{"op": "add", "path": "participants", "value": self.user.username}],
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "id": self.thread.pk,
                "detail": [
                    "You have to be thread owner to add new participants to it."
                ],
            },
        )

    def test_add_empty_username(self):
        """path validates username"""
        ThreadParticipant.objects.set_owner(self.thread, self.user)

        response = self.patch(
            self.api_link, [{"op": "add", "path": "participants", "value": ""}]
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "id": self.thread.pk,
                "detail": ["You have to enter new participant's username."],
            },
        )

    def test_add_nonexistant_user(self):
        """can't user two times"""
        ThreadParticipant.objects.set_owner(self.thread, self.user)

        response = self.patch(
            self.api_link,
            [{"op": "add", "path": "participants", "value": "InvalidUser"}],
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {"id": self.thread.pk, "detail": ["No user with such name exists."]},
        )

    def test_add_already_participant(self):
        """can't add user that is already participant"""
        ThreadParticipant.objects.set_owner(self.thread, self.user)

        response = self.patch(
            self.api_link,
            [{"op": "add", "path": "participants", "value": self.user.username}],
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "id": self.thread.pk,
                "detail": ["This user is already thread participant."],
            },
        )

    def test_add_blocking_user(self):
        """can't add user that is already participant"""
        ThreadParticipant.objects.set_owner(self.thread, self.user)
        self.other_user.blocks.add(self.user)

        response = self.patch(
            self.api_link,
            [{"op": "add", "path": "participants", "value": self.other_user.username}],
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {"id": self.thread.pk, "detail": ["OtherUser is blocking you."]},
        )

    @patch_user_acl(other_user_cant_use_private_threads)
    def test_add_no_perm_user(self):
        """can't add user that has no permission to use private threads"""
        ThreadParticipant.objects.set_owner(self.thread, self.user)

        response = self.patch(
            self.api_link,
            [{"op": "add", "path": "participants", "value": self.other_user.username}],
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "id": self.thread.pk,
                "detail": ["OtherUser can't participate in private threads."],
            },
        )

    @patch_user_acl({"max_private_thread_participants": 3})
    def test_add_too_many_users(self):
        """can't add user that is already participant"""
        ThreadParticipant.objects.set_owner(self.thread, self.user)

        for i in range(3):
            user = create_test_user("User%s" % i, "user%s@example.com" % i)
            ThreadParticipant.objects.add_participants(self.thread, [user])

        response = self.patch(
            self.api_link,
            [{"op": "add", "path": "participants", "value": self.other_user.username}],
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "id": self.thread.pk,
                "detail": ["You can't add any more new users to this thread."],
            },
        )

    def test_add_user_closed_thread(self):
        """adding user to closed thread fails for non-moderator"""
        ThreadParticipant.objects.set_owner(self.thread, self.user)

        self.thread.is_closed = True
        self.thread.save()

        response = self.patch(
            self.api_link,
            [{"op": "add", "path": "participants", "value": self.other_user.username}],
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "id": self.thread.pk,
                "detail": ["Only moderators can add participants to closed threads."],
            },
        )

    def test_add_user(self):
        """
        adding user to thread add user to thread as participant,
        sets event and emails them
        """
        ThreadParticipant.objects.set_owner(self.thread, self.user)

        self.patch(
            self.api_link,
            [{"op": "add", "path": "participants", "value": self.other_user.username}],
        )

        # event was set on thread
        event = self.thread.post_set.order_by("id").last()
        self.assertTrue(event.is_event)
        self.assertTrue(event.event_type, "added_participant")

        # notification about new private thread was sent to other user
        self.assertEqual(len(mail.outbox), 1)
        email = mail.outbox[-1]

        self.assertIn(self.user.username, email.subject)
        self.assertIn(self.thread.title, email.subject)

    @patch_user_acl({"can_moderate_private_threads": True})
    def test_add_user_to_other_user_thread_moderator(self):
        """moderators can add users to other users threads"""
        ThreadParticipant.objects.set_owner(self.thread, self.other_user)

        self.thread.has_reported_posts = True
        self.thread.save()

        self.patch(
            self.api_link,
            [{"op": "add", "path": "participants", "value": self.user.username}],
        )

        # event was set on thread
        event = self.thread.post_set.order_by("id").last()
        self.assertTrue(event.is_event)
        self.assertTrue(event.event_type, "entered_thread")

        # notification about new private thread wasn't send because we invited ourselves
        self.assertEqual(len(mail.outbox), 0)

    @patch_user_acl({"can_moderate_private_threads": True})
    def test_add_user_to_closed_moderator(self):
        """moderators can add users to closed threads"""
        ThreadParticipant.objects.set_owner(self.thread, self.user)

        self.thread.is_closed = True
        self.thread.save()

        self.patch(
            self.api_link,
            [{"op": "add", "path": "participants", "value": self.other_user.username}],
        )

        # event was set on thread
        event = self.thread.post_set.order_by("id").last()
        self.assertTrue(event.is_event)
        self.assertTrue(event.event_type, "added_participant")

        # notification about new private thread was sent to other user
        self.assertEqual(len(mail.outbox), 1)
        email = mail.outbox[-1]

        self.assertIn(self.user.username, email.subject)
        self.assertIn(self.thread.title, email.subject)


class PrivateThreadRemoveParticipantApiTests(PrivateThreadPatchApiTestCase):
    def test_remove_empty(self):
        """api handles empty user id"""
        ThreadParticipant.objects.set_owner(self.thread, self.user)

        response = self.patch(
            self.api_link, [{"op": "remove", "path": "participants", "value": ""}]
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {"id": self.thread.pk, "detail": ["A valid integer is required."]},
        )

    def test_remove_invalid(self):
        """api validates user id type"""
        ThreadParticipant.objects.set_owner(self.thread, self.user)

        response = self.patch(
            self.api_link, [{"op": "remove", "path": "participants", "value": "string"}]
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {"id": self.thread.pk, "detail": ["A valid integer is required."]},
        )

    def test_remove_nonexistant(self):
        """removed user has to be participant"""
        ThreadParticipant.objects.set_owner(self.thread, self.user)

        response = self.patch(
            self.api_link,
            [{"op": "remove", "path": "participants", "value": self.other_user.pk}],
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {"id": self.thread.pk, "detail": ["Participant doesn't exist."]},
        )

    def test_remove_not_owner(self):
        """api validates if user trying to remove other user is an owner"""
        ThreadParticipant.objects.set_owner(self.thread, self.other_user)
        ThreadParticipant.objects.add_participants(self.thread, [self.user])

        response = self.patch(
            self.api_link,
            [{"op": "remove", "path": "participants", "value": self.other_user.pk}],
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "id": self.thread.pk,
                "detail": [
                    "You have to be thread owner to remove participants from it."
                ],
            },
        )

    def test_owner_remove_user_closed_thread(self):
        """api disallows owner to remove other user from closed thread"""
        ThreadParticipant.objects.set_owner(self.thread, self.user)
        ThreadParticipant.objects.add_participants(self.thread, [self.other_user])

        self.thread.is_closed = True
        self.thread.save()

        response = self.patch(
            self.api_link,
            [{"op": "remove", "path": "participants", "value": self.other_user.pk}],
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "id": self.thread.pk,
                "detail": [
                    "Only moderators can remove participants from closed threads."
                ],
            },
        )

    def test_user_leave_thread(self):
        """api allows user to remove themself from thread"""
        ThreadParticipant.objects.set_owner(self.thread, self.other_user)
        ThreadParticipant.objects.add_participants(self.thread, [self.user])

        self.user.subscription_set.create(category=self.category, thread=self.thread)

        response = self.patch(
            self.api_link,
            [{"op": "remove", "path": "participants", "value": self.user.pk}],
        )

        self.assertEqual(response.status_code, 200)
        self.assertFalse(response.json()["deleted"])

        # thread still exists
        self.assertTrue(Thread.objects.get(pk=self.thread.pk))

        # leave event has valid type
        event = self.thread.post_set.order_by("id").last()
        self.assertTrue(event.is_event)
        self.assertTrue(event.event_type, "participant_left")

        # valid users were flagged for sync
        self.user.refresh_from_db()
        self.assertTrue(self.user.sync_unread_private_threads)

        self.other_user.refresh_from_db()
        self.assertTrue(self.other_user.sync_unread_private_threads)

        # user was removed from participation
        self.assertEqual(self.thread.participants.count(), 1)
        self.assertEqual(self.thread.participants.filter(pk=self.user.pk).count(), 0)

        # thread was removed from user subscriptions
        self.assertEqual(self.user.subscription_set.count(), 0)

    def test_user_leave_closed_thread(self):
        """api allows user to remove themself from closed thread"""
        ThreadParticipant.objects.set_owner(self.thread, self.other_user)
        ThreadParticipant.objects.add_participants(self.thread, [self.user])

        self.thread.is_closed = True
        self.thread.save()

        response = self.patch(
            self.api_link,
            [{"op": "remove", "path": "participants", "value": self.user.pk}],
        )

        self.assertEqual(response.status_code, 200)
        self.assertFalse(response.json()["deleted"])

        # thread still exists
        self.assertTrue(Thread.objects.get(pk=self.thread.pk))

        # leave event has valid type
        event = self.thread.post_set.order_by("id").last()
        self.assertTrue(event.is_event)
        self.assertTrue(event.event_type, "participant_left")

        # valid users were flagged for sync
        self.user.refresh_from_db()
        self.assertTrue(self.user.sync_unread_private_threads)

        self.other_user.refresh_from_db()
        self.assertTrue(self.other_user.sync_unread_private_threads)

        # user was removed from participation
        self.assertEqual(self.thread.participants.count(), 1)
        self.assertEqual(self.thread.participants.filter(pk=self.user.pk).count(), 0)

    @patch_user_acl({"can_moderate_private_threads": True})
    def test_moderator_remove_user(self):
        """api allows moderator to remove other user"""
        removed_user = create_test_user("RemovedUser", "removeduser@example.com")

        ThreadParticipant.objects.set_owner(self.thread, self.other_user)
        ThreadParticipant.objects.add_participants(
            self.thread, [self.user, removed_user]
        )

        response = self.patch(
            self.api_link,
            [{"op": "remove", "path": "participants", "value": removed_user.pk}],
        )

        self.assertEqual(response.status_code, 200)
        self.assertFalse(response.json()["deleted"])

        # thread still exists
        self.assertTrue(Thread.objects.get(pk=self.thread.pk))

        # leave event has valid type
        event = self.thread.post_set.order_by("id").last()
        self.assertTrue(event.is_event)
        self.assertTrue(event.event_type, "participant_removed")

        # valid users were flagged for sync
        self.user.refresh_from_db()
        self.assertTrue(self.user.sync_unread_private_threads)

        self.other_user.refresh_from_db()
        self.assertTrue(self.other_user.sync_unread_private_threads)

        removed_user.refresh_from_db()
        self.assertTrue(removed_user.sync_unread_private_threads)

        # user was removed from participation
        self.assertEqual(self.thread.participants.count(), 2)
        self.assertEqual(self.thread.participants.filter(pk=removed_user.pk).count(), 0)

    def test_owner_remove_user(self):
        """api allows owner to remove other user"""
        ThreadParticipant.objects.set_owner(self.thread, self.user)
        ThreadParticipant.objects.add_participants(self.thread, [self.other_user])

        response = self.patch(
            self.api_link,
            [{"op": "remove", "path": "participants", "value": self.other_user.pk}],
        )

        self.assertEqual(response.status_code, 200)
        self.assertFalse(response.json()["deleted"])

        # thread still exists
        self.assertTrue(Thread.objects.get(pk=self.thread.pk))

        # leave event has valid type
        event = self.thread.post_set.order_by("id").last()
        self.assertTrue(event.is_event)
        self.assertTrue(event.event_type, "participant_removed")

        # valid users were flagged for sync
        self.user.refresh_from_db()
        self.assertTrue(self.user.sync_unread_private_threads)

        self.other_user.refresh_from_db()
        self.assertTrue(self.other_user.sync_unread_private_threads)

        # user was removed from participation
        self.assertEqual(self.thread.participants.count(), 1)
        self.assertEqual(
            self.thread.participants.filter(pk=self.other_user.pk).count(), 0
        )

    def test_owner_leave_thread(self):
        """api allows owner to remove hisemf from thread, causing thread to close"""
        ThreadParticipant.objects.set_owner(self.thread, self.user)
        ThreadParticipant.objects.add_participants(self.thread, [self.other_user])

        response = self.patch(
            self.api_link,
            [{"op": "remove", "path": "participants", "value": self.user.pk}],
        )

        self.assertEqual(response.status_code, 200)
        self.assertFalse(response.json()["deleted"])

        # thread still exists and is closed
        self.assertTrue(Thread.objects.get(pk=self.thread.pk).is_closed)

        # leave event has valid type
        event = self.thread.post_set.order_by("id").last()
        self.assertTrue(event.is_event)
        self.assertTrue(event.event_type, "owner_left")

        # valid users were flagged for sync
        self.user.refresh_from_db()
        self.assertTrue(self.user.sync_unread_private_threads)

        self.other_user.refresh_from_db()
        self.assertTrue(self.other_user.sync_unread_private_threads)

        # user was removed from participation
        self.assertEqual(self.thread.participants.count(), 1)
        self.assertEqual(self.thread.participants.filter(pk=self.user.pk).count(), 0)

    def test_last_user_leave_thread(self):
        """api allows last user leave thread, causing thread to delete"""
        ThreadParticipant.objects.set_owner(self.thread, self.user)

        response = self.patch(
            self.api_link,
            [{"op": "remove", "path": "participants", "value": self.user.pk}],
        )

        self.assertEqual(response.status_code, 200)
        self.assertTrue(response.json()["deleted"])

        # thread is gone
        with self.assertRaises(Thread.DoesNotExist):
            Thread.objects.get(pk=self.thread.pk)

        # valid users were flagged for sync
        self.user.refresh_from_db()
        self.assertTrue(self.user.sync_unread_private_threads)


class PrivateThreadTakeOverApiTests(PrivateThreadPatchApiTestCase):
    def test_empty_user_id(self):
        """api handles empty user id"""
        ThreadParticipant.objects.set_owner(self.thread, self.user)

        response = self.patch(
            self.api_link, [{"op": "replace", "path": "owner", "value": ""}]
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {"id": self.thread.pk, "detail": ["A valid integer is required."]},
        )

    def test_invalid_user_id(self):
        """api handles invalid user id"""
        ThreadParticipant.objects.set_owner(self.thread, self.user)

        response = self.patch(
            self.api_link, [{"op": "replace", "path": "owner", "value": "dsadsa"}]
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {"id": self.thread.pk, "detail": ["A valid integer is required."]},
        )

    def test_nonexistant_user_id(self):
        """api handles nonexistant user id"""
        ThreadParticipant.objects.set_owner(self.thread, self.user)

        response = self.patch(
            self.api_link,
            [{"op": "replace", "path": "owner", "value": self.other_user.pk}],
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {"id": self.thread.pk, "detail": ["Participant doesn't exist."]},
        )

    def test_no_permission(self):
        """non-moderator/owner can't change owner"""
        ThreadParticipant.objects.set_owner(self.thread, self.other_user)
        ThreadParticipant.objects.add_participants(self.thread, [self.user])

        response = self.patch(
            self.api_link, [{"op": "replace", "path": "owner", "value": self.user.pk}]
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "id": self.thread.pk,
                "detail": [
                    "Only thread owner and moderators can change threads owners."
                ],
            },
        )

    def test_no_change(self):
        """api validates that new owner id is same as current owner"""
        ThreadParticipant.objects.set_owner(self.thread, self.user)
        ThreadParticipant.objects.add_participants(self.thread, [self.other_user])

        response = self.patch(
            self.api_link, [{"op": "replace", "path": "owner", "value": self.user.pk}]
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {"id": self.thread.pk, "detail": ["This user already is thread owner."]},
        )

    def test_change_closed_thread_owner(self):
        """non-moderator can't change owner in closed thread"""
        ThreadParticipant.objects.set_owner(self.thread, self.user)
        ThreadParticipant.objects.add_participants(self.thread, [self.other_user])

        self.thread.is_closed = True
        self.thread.save()

        response = self.patch(
            self.api_link,
            [{"op": "replace", "path": "owner", "value": self.other_user.pk}],
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "id": self.thread.pk,
                "detail": ["Only moderators can change closed threads owners."],
            },
        )

    def test_owner_change_thread_owner(self):
        """owner can pass thread ownership to other participant"""
        ThreadParticipant.objects.set_owner(self.thread, self.user)
        ThreadParticipant.objects.add_participants(self.thread, [self.other_user])

        response = self.patch(
            self.api_link,
            [{"op": "replace", "path": "owner", "value": self.other_user.pk}],
        )

        self.assertEqual(response.status_code, 200)

        # valid users were flagged for sync
        self.user.refresh_from_db()
        self.assertFalse(self.user.sync_unread_private_threads)

        self.other_user.refresh_from_db()
        self.assertTrue(self.other_user.sync_unread_private_threads)

        # ownership was transfered
        self.assertEqual(self.thread.participants.count(), 2)
        self.assertTrue(ThreadParticipant.objects.get(user=self.other_user).is_owner)
        self.assertFalse(ThreadParticipant.objects.get(user=self.user).is_owner)

        # change was recorded in event
        event = self.thread.post_set.order_by("id").last()
        self.assertTrue(event.is_event)
        self.assertTrue(event.event_type, "changed_owner")

    @patch_user_acl({"can_moderate_private_threads": True})
    def test_moderator_change_owner(self):
        """moderator can change thread owner to other user"""
        new_owner = create_test_user("NewOwner", "newowner@example.com")

        ThreadParticipant.objects.set_owner(self.thread, self.other_user)
        ThreadParticipant.objects.add_participants(self.thread, [self.user, new_owner])

        response = self.patch(
            self.api_link, [{"op": "replace", "path": "owner", "value": new_owner.pk}]
        )

        self.assertEqual(response.status_code, 200)

        # valid users were flagged for sync
        new_owner.refresh_from_db()
        self.assertTrue(new_owner.sync_unread_private_threads)

        self.user.refresh_from_db()
        self.assertFalse(self.user.sync_unread_private_threads)

        self.other_user.refresh_from_db()
        self.assertTrue(self.other_user.sync_unread_private_threads)

        # ownership was transferred
        self.assertEqual(self.thread.participants.count(), 3)
        self.assertTrue(ThreadParticipant.objects.get(user=new_owner).is_owner)
        self.assertFalse(ThreadParticipant.objects.get(user=self.user).is_owner)
        self.assertFalse(ThreadParticipant.objects.get(user=self.other_user).is_owner)

        # change was recorded in event
        event = self.thread.post_set.order_by("id").last()
        self.assertTrue(event.is_event)
        self.assertTrue(event.event_type, "changed_owner")

    @patch_user_acl({"can_moderate_private_threads": True})
    def test_moderator_takeover(self):
        """moderator can takeover the thread"""
        ThreadParticipant.objects.set_owner(self.thread, self.other_user)
        ThreadParticipant.objects.add_participants(self.thread, [self.user])

        response = self.patch(
            self.api_link, [{"op": "replace", "path": "owner", "value": self.user.pk}]
        )

        self.assertEqual(response.status_code, 200)

        # valid users were flagged for sync
        self.user.refresh_from_db()
        self.assertFalse(self.user.sync_unread_private_threads)

        self.other_user.refresh_from_db()
        self.assertTrue(self.other_user.sync_unread_private_threads)

        # ownership was transfered
        self.assertEqual(self.thread.participants.count(), 2)
        self.assertTrue(ThreadParticipant.objects.get(user=self.user).is_owner)
        self.assertFalse(ThreadParticipant.objects.get(user=self.other_user).is_owner)

        # change was recorded in event
        event = self.thread.post_set.order_by("id").last()
        self.assertTrue(event.is_event)
        self.assertTrue(event.event_type, "tookover")

    @patch_user_acl({"can_moderate_private_threads": True})
    def test_moderator_closed_thread_takeover(self):
        """moderator can takeover closed thread thread"""
        ThreadParticipant.objects.set_owner(self.thread, self.other_user)
        ThreadParticipant.objects.add_participants(self.thread, [self.user])

        self.thread.is_closed = True
        self.thread.save()

        response = self.patch(
            self.api_link, [{"op": "replace", "path": "owner", "value": self.user.pk}]
        )

        self.assertEqual(response.status_code, 200)

        # valid users were flagged for sync
        self.user.refresh_from_db()
        self.assertFalse(self.user.sync_unread_private_threads)

        self.other_user.refresh_from_db()
        self.assertTrue(self.other_user.sync_unread_private_threads)

        # ownership was transferred
        self.assertEqual(self.thread.participants.count(), 2)
        self.assertTrue(ThreadParticipant.objects.get(user=self.user).is_owner)
        self.assertFalse(ThreadParticipant.objects.get(user=self.other_user).is_owner)

        # change was recorded in event
        event = self.thread.post_set.order_by("id").last()
        self.assertTrue(event.is_event)
        self.assertTrue(event.event_type, "tookover")



# Generated by Django 2.2.1 on 2019-05-19 00:16

from django.db import migrations

from ..hydrators import dehydrate_value

settings = [
    {
        "setting": "enable_sso",
        "python_type": "bool",
        "dry_value": False,
        "is_public": True,
    },
    {"setting": "sso_public_key", "is_public": False},
    {"setting": "sso_private_key", "is_public": False},
    {"setting": "sso_url", "is_public": False},
]


def create_settings(apps, _):
    Setting = apps.get_model("misago_conf", "Setting")
    for setting in settings:
        data = setting.copy()
        if "python_type" in data and "dry_value" in data:
            data["dry_value"] = dehydrate_value(data["python_type"], data["dry_value"])

        Setting.objects.create(**setting)


class Migration(migrations.Migration):

    dependencies = [("misago_conf", "0004_create_settings")]

    operations = [migrations.RunPython(create_settings)]



from django.urls import reverse

api_link = reverse("misago:api:parse-markup")


def test_api_rejects_unauthenticated_user(db, client):
    response = client.post(api_link)
    assert response.status_code == 403


def test_api_rejects_request_without_data(user_client):
    response = user_client.post(api_link)
    assert response.status_code == 400
    assert response.json() == {"detail": "You have to enter a message."}


def test_api_rejects_request_with_invalid_shaped_data(user_client):
    response = user_client.post(api_link, "[]", content_type="application/json")
    assert response.status_code == 400
    assert response.json() == {
        "detail": "Invalid data. Expected a dictionary, but got list."
    }

    response = user_client.post(api_link, "123", content_type="application/json")
    assert response.status_code == 400
    assert response.json() == {
        "detail": "Invalid data. Expected a dictionary, but got int."
    }

    response = user_client.post(api_link, '"string"', content_type="application/json")
    assert response.status_code == 400
    assert response.json() == {
        "detail": "Invalid data. Expected a dictionary, but got str."
    }


def test_api_rejects_request_with_malformed_data(user_client):
    response = user_client.post(api_link, "malformed", content_type="application/json")
    assert response.status_code == 400
    assert response.json() == {
        "detail": "JSON parse error - Expecting value: line 1 column 1 (char 0)"
    }


def test_api_validates_that_post_has_content(user_client):
    response = user_client.post(api_link, json={"post": ""})
    assert response.status_code == 400
    assert response.json() == {"detail": "You have to enter a message."}


# regression test for #929
def test_api_strips_whitespace_from_post_before_validating_length(user_client):
    response = user_client.post(api_link, json={"post": "\n"})
    assert response.status_code == 400
    assert response.json() == {"detail": "You have to enter a message."}


def test_api_casts_post_value_to_string(user_client):
    response = user_client.post(api_link, json={"post": 123})
    assert response.status_code == 400
    assert response.json() == {
        "detail": "Posted message should be at least 5 characters long (it has 3)."
    }


def test_api_returns_parsed_value(user_client):
    response = user_client.post(api_link, json={"post": "Hello world!"})
    assert response.status_code == 200
    assert response.json() == {"parsed": "<p>Hello world!</p>"}



from io import StringIO

from django.core.management import call_command
from django.test import TestCase

from .. import test
from ...categories.models import Category
from ..management.commands import updatepostschecksums
from ..models import Post


class UpdatePostsChecksumsTests(TestCase):
    def test_no_posts_to_update(self):
        """command works when there are no posts"""
        command = updatepostschecksums.Command()

        out = StringIO()
        call_command(command, stdout=out)
        command_output = out.getvalue().strip()

        self.assertEqual(command_output, "No posts were found")

    def test_posts_update(self):
        """command updates posts checksums"""
        category = Category.objects.all_categories()[:1][0]

        threads = [test.post_thread(category) for _ in range(5)]
        for _, thread in enumerate(threads):
            [test.reply_thread(thread) for _ in range(3)]
            thread.save()

        Post.objects.update(parsed="Hello world!")
        for post in Post.objects.all():
            self.assertFalse(post.is_valid)

        command = updatepostschecksums.Command()

        out = StringIO()
        call_command(command, stdout=out)

        command_output = out.getvalue().splitlines()[-1].strip()
        self.assertEqual(command_output, "Updated 20 posts checksums")

        for post in Post.objects.all():
            self.assertTrue(post.is_valid)



import re

from django.utils.translation import gettext as _

QUOTE_HEADER_RE = re.compile(
    r"""
<div class="quote-heading">(?P<title>.*?)</div>
""".strip(),
    re.IGNORECASE | re.MULTILINE | re.DOTALL,
)

SPOILER_REVEAL_BTN = '<button class="spoiler-reveal" type="button"></button>'


def finalize_markup(post):
    post = QUOTE_HEADER_RE.sub(replace_quote_headers, post)
    post = replace_spoiler_reveal_buttons(post)
    return post


def replace_quote_headers(matchobj):
    title = matchobj.group("title")
    if title:
        quote_title = _("%(title)s has written:") % {"title": title}
    else:
        quote_title = _("Quoted message:")
    return '<div class="quote-heading">%s</div>' % quote_title


def replace_spoiler_headers(matchobj):
    title = matchobj.group("title")
    if title:
        spoiler_title = _("%(title)s:") % {"title": title}
    else:
        spoiler_title = _("Spoiler:")
    return '<div class="spoiler-heading">%s</div>' % spoiler_title


def replace_spoiler_reveal_buttons(post):
    final_btn = SPOILER_REVEAL_BTN.replace("></", ">%s</" % _("Reveal spoiler"))
    return post.replace(SPOILER_REVEAL_BTN, final_btn)



from io import BytesIO

import pytest
from PIL import Image

from django.core.files.uploadedfile import SimpleUploadedFile
from django.urls import reverse

from ... import settings
from ...models import Setting

admin_link = reverse("misago:admin:settings:users:index")


def create_image(width, height):
    image = Image.new("RGBA", (width, height))
    stream = BytesIO()
    image.save(stream, "PNG")
    stream.seek(0)
    return SimpleUploadedFile("image.png", stream.read(), "image/jpeg")


def submit_image(admin_client, image=""):
    data = {
        "account_activation": "user",
        "username_length_min": 10,
        "username_length_max": 10,
        "anonymous_username": "Deleted",
        "avatar_upload_limit": 2000,
        "default_avatar": "gravatar",
        "default_gravatar_fallback": "dynamic",
        "signature_length_max": 100,
        "blank_avatar": image,
        "subscribe_start": "no",
        "subscribe_reply": "no",
        "users_per_page": 12,
        "users_per_page_orphans": 4,
        "top_posters_ranking_length": 10,
        "top_posters_ranking_size": 10,
        "allow_data_downloads": "no",
        "data_downloads_expiration": 48,
        "allow_delete_own_account": "no",
        "new_inactive_accounts_delete": 0,
        "ip_storage_time": 0,
    }

    return admin_client.post(admin_link, data)


@pytest.fixture
def setting(db):
    return Setting.objects.get(setting="blank_avatar")


@pytest.fixture
def setting_with_value(admin_client, setting):
    min_size = max(settings.MISAGO_AVATARS_SIZES)
    image_file = create_image(min_size, min_size)
    submit_image(admin_client, image_file)

    setting.refresh_from_db()
    return setting


def test_uploaded_image_is_rejected_if_its_not_square(admin_client, setting):
    image_file = create_image(100, 200)
    submit_image(admin_client, image_file)

    setting.refresh_from_db()
    assert not setting.value


def test_uploaded_image_is_rejected_if_its_smaller_than_max_avatar_size(
    admin_client, setting
):
    min_size = max(settings.MISAGO_AVATARS_SIZES)
    image_file = create_image(min_size - 1, min_size - 1)
    submit_image(admin_client, image_file)

    setting.refresh_from_db()
    assert not setting.value


def test_valid_blank_avatar_can_be_uploaded(admin_client, setting):
    min_size = max(settings.MISAGO_AVATARS_SIZES)
    image_file = create_image(min_size, min_size)
    submit_image(admin_client, image_file)

    setting.refresh_from_db()
    assert setting.value


def test_submitting_form_without_new_image_doesnt_unset_existing_image(
    admin_client, setting_with_value
):
    submit_image(admin_client)
    setting_with_value.refresh_from_db()
    assert setting_with_value.value



from unittest.mock import Mock
from urllib.parse import parse_qsl, urlparse

from ...conf.test import override_dynamic_settings
from ..client import SESSION_STATE, create_login_url


@override_dynamic_settings(
    enable_oauth2_client=True,
    oauth2_client_id="clientid123",
    oauth2_scopes="some scopes",
    oauth2_login_url="https://example.com/oauth2/login",
)
def test_oauth2_login_url_is_created(dynamic_settings):
    request = Mock(
        session={},
        settings=dynamic_settings,
        build_absolute_uri=lambda url: f"http://mysite.com{url or ''}",
    )

    login_url = create_login_url(request)

    # State set in session?
    assert request.session.get(SESSION_STATE)

    # Redirect url is valid?
    redirect_to = urlparse(login_url)
    assert redirect_to.netloc == "example.com"
    assert redirect_to.path == "/oauth2/login"
    assert parse_qsl(redirect_to.query) == [
        ("response_type", "code"),
        ("client_id", "clientid123"),
        ("redirect_uri", "http://mysite.com/oauth2/complete/"),
        ("scope", "some scopes"),
        ("state", request.session[SESSION_STATE]),
    ]



import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("misago_threads", "0001_initial"),
        ("misago_categories", "0003_categories_roles"),
    ]

    operations = [
        migrations.AddField(
            model_name="category",
            name="last_thread",
            field=models.ForeignKey(
                related_name="+",
                on_delete=django.db.models.deletion.SET_NULL,
                blank=True,
                to="misago_threads.Thread",
                null=True,
            ),
            preserve_default=True,
        )
    ]



from django.urls import reverse

from .pages import user_profile, usercp, users_list
from .serializers import AnonymousUserSerializer, AuthenticatedUserSerializer


def user_links(request):
    if request.include_frontend_context:
        request.frontend_context.update(
            {
                "REQUEST_ACTIVATION_URL": reverse("misago:request-activation"),
                "FORGOTTEN_PASSWORD_URL": reverse("misago:forgotten-password"),
                "BANNED_URL": reverse("misago:banned"),
                "USERCP_URL": reverse("misago:options"),
                "USERS_LIST_URL": reverse("misago:users"),
                "AUTH_API": reverse("misago:api:auth"),
                "AUTH_CRITERIA_API": reverse("misago:api:auth-criteria"),
                "USERS_API": reverse("misago:api:user-list"),
                "CAPTCHA_API": reverse("misago:api:captcha-question"),
                "USERNAME_CHANGES_API": reverse("misago:api:usernamechange-list"),
                "MENTION_API": reverse("misago:api:mention-suggestions"),
            }
        )

    return {
        "USERCP_URL": usercp.get_default_link(),
        "USERS_LIST_URL": users_list.get_default_link(),
        "USER_PROFILE_URL": user_profile.get_default_link(),
    }


def preload_user_json(request):
    if not request.include_frontend_context:
        return {}

    request.frontend_context.update(
        {"isAuthenticated": bool(request.user.is_authenticated)}
    )

    if request.user.is_authenticated:
        serializer = AuthenticatedUserSerializer
    else:
        serializer = AnonymousUserSerializer

    serialized_user = serializer(request.user, context={"acl": request.user_acl}).data
    request.frontend_context.update({"user": serialized_user})

    return {}



from io import BytesIO

import pytest
from PIL import Image
from django.urls import reverse


@pytest.fixture
def admin_link():
    return reverse("misago:admin:settings:icons:index")


def get_image(width, height):
    image = Image.new("RGBA", (width, height))
    buffer = BytesIO()
    image.save(buffer, "PNG")
    buffer.seek(0)
    return buffer.read()


@pytest.fixture
def image():
    return get_image(185, 185)


@pytest.fixture
def image_alt():
    return get_image(85, 85)


@pytest.fixture
def image_non_square():
    return get_image(180, 200)


@pytest.fixture
def image_small():
    return get_image(20, 20)



from unittest.mock import Mock, patch

import pytest
from django.contrib.auth import get_user_model

from ...conf.test import override_dynamic_settings
from ..exceptions import OAuth2UserDataValidationError
from ..models import Subject
from ..user import get_user_from_data

User = get_user_model()


def test_activated_user_is_created_from_valid_data(db, dynamic_settings):
    user, created = get_user_from_data(
        Mock(settings=dynamic_settings, user_ip="83.0.0.1"),
        {
            "id": "1234",
            "name": "NewUser",
            "email": "user@example.com",
            "avatar": None,
        },
        {},
    )

    assert created
    assert user.id
    assert user.username == "NewUser"
    assert user.slug == "newuser"
    assert user.email == "user@example.com"
    assert user.requires_activation == User.ACTIVATION_NONE

    user_by_name = User.objects.get_by_username("NewUser")
    assert user_by_name.id == user.id

    user_by_email = User.objects.get_by_email("user@example.com")
    assert user_by_email.id == user.id


def test_user_subject_is_created_from_valid_data(db, dynamic_settings):
    user, created = get_user_from_data(
        Mock(settings=dynamic_settings, user_ip="83.0.0.1"),
        {
            "id": "1234",
            "name": "NewUser",
            "email": "user@example.com",
            "avatar": None,
        },
        {},
    )

    assert created
    assert user

    user_subject = Subject.objects.get(sub="1234")
    assert user_subject.user_id == user.id


def test_user_is_created_with_avatar_from_valid_data(db, dynamic_settings):
    user, created = get_user_from_data(
        Mock(settings=dynamic_settings, user_ip="83.0.0.1"),
        {
            "id": "1234",
            "name": "NewUser",
            "email": "user@example.com",
            "avatar": "https://placekitten.com/600/500",
        },
        {},
    )

    assert created
    assert user
    assert user.avatars
    assert user.avatar_set.exists()


@override_dynamic_settings(account_activation="admin")
def test_user_is_created_with_admin_activation_from_valid_data(db, dynamic_settings):
    user, created = get_user_from_data(
        Mock(settings=dynamic_settings, user_ip="83.0.0.1"),
        {
            "id": "1234",
            "name": "NewUser",
            "email": "user@example.com",
            "avatar": None,
        },
        {},
    )

    assert created
    assert user
    assert user.requires_activation == User.ACTIVATION_ADMIN


def user_noop_filter(*args):
    pass


def test_user_name_conflict_during_creation_from_valid_data_is_handled(
    user, dynamic_settings
):
    with pytest.raises(OAuth2UserDataValidationError) as excinfo:
        # Custom filters disable build in filters
        with patch(
            "misago.oauth2.validation.oauth2_user_data_filters",
            [user_noop_filter],
        ):
            get_user_from_data(
                Mock(settings=dynamic_settings, user_ip="83.0.0.1"),
                {
                    "id": "1234",
                    "name": user.username,
                    "email": "test@example.com",
                    "avatar": None,
                },
                {},
            )

    assert excinfo.value.error_list == ["This username is not available."]


def test_user_email_conflict_during_creation_from_valid_data_is_handled(
    user, dynamic_settings
):
    with pytest.raises(OAuth2UserDataValidationError) as excinfo:
        get_user_from_data(
            Mock(settings=dynamic_settings, user_ip="83.0.0.1"),
            {
                "id": "1234",
                "name": "NewUser",
                "email": user.email,
                "avatar": None,
            },
            {},
        )

    assert excinfo.value.error_list == ["This e-mail address is not available."]



from ..context_processors import icons
from ..models import Icon


def test_context_processor_adds_icons_key_to_context(db):
    assert "icons" in icons(None)


def test_icons_context_defaults_to_empty_dict(db):
    assert icons(None) == {"icons": {}}


def test_set_favicon_icon_is_present_in_context(favicon):
    assert Icon.TYPE_FAVICON in icons(None)["icons"]


def test_set_favicon_32_icon_is_present_in_context(favicon_32):
    assert Icon.TYPE_FAVICON_32 in icons(None)["icons"]


def test_set_favicon_16_icon_is_present_in_context(favicon_16):
    assert Icon.TYPE_FAVICON_16 in icons(None)["icons"]


def test_set_apple_touch_icon_icon_is_present_in_context(apple_touch_icon):
    assert Icon.TYPE_APPLE_TOUCH_ICON in icons(None)["icons"]



from xml.etree.ElementTree import Element

import markdown
from markdown.inlinepatterns import LinkInlineProcessor

IMAGES_RE = r"\!\((<.*?>|([^\)]*))\)"


class ShortImagesExtension(markdown.Extension):
    def extendMarkdown(self, md):
        md.registerExtension(self)
        md.inlinePatterns.register(
            ShortImagePattern(IMAGES_RE, md), "misago_short_images", 200
        )


class ShortImagePattern(LinkInlineProcessor):
    def handleMatch(self, m, _):
        img_src = m.groups()[1].strip()
        if not img_src:
            return None, None, None

        img = Element("img")
        img.set("src", img_src)
        img.set("alt", img_src)
        return img, m.start(0), m.end(0)



from datetime import timedelta

from django.conf import settings
from django.contrib.auth import get_user_model
from django.core.cache import cache
from django.utils import timezone

from . import render
from ...threads.models import Post, Thread, Attachment
from ...users.models import DataDownload

VERSION_CHECK_CACHE_KEY = "misago_version_check"

User = get_user_model()


def admin_index(request):
    totals = count_db_items()
    checks = {
        "address": check_forum_address(request),
        "cache": check_cache(),
        "data_downloads": check_data_downloads(),
        "debug": check_debug_status(),
        "https": check_https(request),
        "inactive_users": check_inactive_users(),
    }

    return render(
        request,
        "misago/admin/dashboard/index.html",
        {"totals": totals, "checks": checks},
    )


def check_cache():
    cache.set("misago_cache_test", "ok")
    return {"is_ok": cache.get("misago_cache_test") == "ok"}


def check_debug_status():
    return {"is_ok": not settings.DEBUG}


def check_https(request):
    return {"is_ok": request.is_secure()}


def check_forum_address(request):
    set_address = request.settings.forum_address
    correct_address = request.build_absolute_uri("/")

    return {
        "is_ok": set_address == correct_address,
        "set_address": set_address,
        "correct_address": correct_address,
    }


def check_data_downloads():
    cutoff = timezone.now() - timedelta(days=3)
    unprocessed_count = DataDownload.objects.filter(
        status__lte=DataDownload.STATUS_PROCESSING, requested_on__lte=cutoff
    ).count()

    return {"is_ok": unprocessed_count == 0, "count": unprocessed_count}


def check_inactive_users():
    count = User.objects.exclude(requires_activation=User.ACTIVATION_NONE).count()
    return {"is_ok": count <= 10, "count": count}


def count_db_items():
    return {
        "attachments": Attachment.objects.count(),
        "threads": Thread.objects.count(),
        "posts": Post.objects.count(),
        "users": User.objects.count(),
    }



from django.urls import reverse

from ...acl.test import patch_user_acl
from ...categories.models import Category
from ...threads.test import post_thread
from ..activepostersranking import build_active_posters_ranking
from ..models import Rank
from ..test import AuthenticatedUserTestCase, create_test_user


class UsersListTestCase(AuthenticatedUserTestCase):
    pass


class UsersListLanderTests(UsersListTestCase):
    @patch_user_acl({"can_browse_users_list": 0})
    def test_lander_no_permission(self):
        """lander returns 403 if user has no permission"""
        response = self.client.get(reverse("misago:users"))
        self.assertEqual(response.status_code, 403)

    def test_lander_redirect(self):
        """lander returns redirect to valid page if user has permission"""
        response = self.client.get(reverse("misago:users"))
        self.assertEqual(response.status_code, 302)
        self.assertTrue(
            response["location"].endswith(reverse("misago:users-active-posters"))
        )


class ActivePostersTests(UsersListTestCase):
    def test_empty_active_posters_list(self):
        """empty active posters page has no showstoppers"""
        view_link = reverse("misago:users-active-posters")

        response = self.client.get(view_link)
        self.assertEqual(response.status_code, 200)

    def test_active_posters_list(self):
        """active posters page has no showstoppers"""
        category = Category.objects.get(slug="first-category")
        view_link = reverse("misago:users-active-posters")

        response = self.client.get(view_link)
        self.assertEqual(response.status_code, 200)

        # Create 50 test users and see if errors appeared
        for i in range(50):
            user = create_test_user("Bob%s" % i, "m%s@te.com" % i, posts=12345)
            post_thread(category, poster=user)

        build_active_posters_ranking()

        response = self.client.get(view_link)
        self.assertEqual(response.status_code, 200)


class UsersRankTests(UsersListTestCase):
    def test_ranks(self):
        """ranks lists are handled correctly"""
        rank_user = create_test_user("Visible", "visible@te.com")

        for rank in Rank.objects.iterator():
            rank_user.rank = rank
            rank_user.save()

            rank_link = reverse("misago:users-rank", kwargs={"slug": rank.slug})
            response = self.client.get(rank_link)

            if rank.is_tab:
                self.assertEqual(response.status_code, 200)
                self.assertContains(response, rank_user.get_absolute_url())
            else:
                self.assertEqual(response.status_code, 404)

    def test_disabled_users(self):
        """ranks lists excludes disabled accounts"""
        rank_user = create_test_user("Visible", "visible@te.com", is_active=False)

        for rank in Rank.objects.iterator():
            rank_user.rank = rank
            rank_user.save()

            rank_link = reverse("misago:users-rank", kwargs={"slug": rank.slug})
            response = self.client.get(rank_link)

            if rank.is_tab:
                self.assertEqual(response.status_code, 200)
                self.assertNotContains(response, rank_user.get_absolute_url())
            else:
                self.assertEqual(response.status_code, 404)

    def test_staff_see_disabled_users(self):
        """ranks lists shows disabled accounts for staff members"""
        self.user.is_staff = True
        self.user.save()

        rank_user = create_test_user("Visible", "visible@te.com", is_active=False)

        for rank in Rank.objects.iterator():
            rank_user.rank = rank
            rank_user.save()

            rank_link = reverse("misago:users-rank", kwargs={"slug": rank.slug})
            response = self.client.get(rank_link)

            if rank.is_tab:
                self.assertEqual(response.status_code, 200)
                self.assertContains(response, rank_user.get_absolute_url())
            else:
                self.assertEqual(response.status_code, 404)



from django.urls import reverse

from ...test import assert_contains, assert_not_contains


def test_body_has_misago_anonymous_css_class_if_user_is_anonymous(db, client):
    response = client.get(reverse("misago:index"))
    assert_contains(response, 'class="misago-anonymous')
    assert_not_contains(response, 'class="misago-authenticated')


def test_body_has_misago_authenticated_css_class_if_user_is_authenticated(user_client):
    response = user_client.get(reverse("misago:index"))
    assert_contains(response, 'class="misago-authenticated')
    assert_not_contains(response, 'class="misago-anonymous')



import json
from unittest.mock import Mock

import pytest
from django.forms import ValidationError
from requests.exceptions import Timeout

from ...conf.test import override_dynamic_settings
from ..validators import validate_with_sfs

cleaned_data = {"email": "test@test.com"}


@pytest.fixture
def request_mock(dynamic_settings):
    return Mock(settings=dynamic_settings, user_ip="127.0.0.1")


@pytest.fixture
def api_mock(mocker):
    return mocker.patch(
        "misago.users.validators.requests",
        get=Mock(
            return_value=Mock(
                content=json.dumps(
                    {"email": {"confidence": 55}, "ip": {"confidence": 45}}
                )
            )
        ),
    )


@override_dynamic_settings(enable_stop_forum_spam=False)
def test_api_is_not_called_if_sfs_is_disabled(api_mock, request_mock):
    validate_with_sfs(request_mock, cleaned_data, None)
    api_mock.get.assert_not_called()


@override_dynamic_settings(enable_stop_forum_spam=True)
def test_api_is_not_called_if_email_is_not_available(api_mock, request_mock):
    validate_with_sfs(request_mock, {}, None)
    api_mock.get.assert_not_called()


@override_dynamic_settings(enable_stop_forum_spam=True, stop_forum_spam_confidence=90)
def test_api_is_called_if_sfs_is_enabled_and_email_is_provided(api_mock, request_mock):
    validate_with_sfs(request_mock, cleaned_data, None)
    api_mock.get.assert_called_once()


@override_dynamic_settings(enable_stop_forum_spam=True, stop_forum_spam_confidence=50)
def test_validator_raises_error_if_ip_score_is_greater_than_confidence(
    api_mock, request_mock
):
    with pytest.raises(ValidationError):
        validate_with_sfs(request_mock, cleaned_data, None)


@override_dynamic_settings(enable_stop_forum_spam=True, stop_forum_spam_confidence=52)
def test_validator_raises_error_if_email_score_is_greater_than_confidence(
    api_mock, request_mock
):
    with pytest.raises(ValidationError):
        validate_with_sfs(request_mock, cleaned_data, None)


@override_dynamic_settings(enable_stop_forum_spam=True)
def test_validator_handles_api_error(mocker, request_mock):
    failing_api_mock = mocker.patch(
        "misago.users.validators.requests",
        get=Mock(return_value=Mock(raise_for_status=Mock(side_effect=Timeout()))),
    )

    validate_with_sfs(request_mock, cleaned_data, None)
    failing_api_mock.get.assert_called_once()


@override_dynamic_settings(enable_stop_forum_spam=True)
def test_validator_logs_api_error(mocker, request_mock):
    failing_api_mock = mocker.patch(
        "misago.users.validators.requests",
        get=Mock(return_value=Mock(raise_for_status=Mock(side_effect=Timeout()))),
    )
    logger_mock = mocker.patch("misago.users.validators.logger", exception=Mock())

    validate_with_sfs(request_mock, cleaned_data, None)
    failing_api_mock.get.assert_called_once()
    logger_mock.exception.assert_called_once()


@override_dynamic_settings(enable_stop_forum_spam=True)
def test_validator_handles_malformed_api_response(mocker, request_mock):
    failing_api_mock = mocker.patch(
        "misago.users.validators.requests", get=Mock(return_value=Mock(content="{}"))
    )

    validate_with_sfs(request_mock, cleaned_data, None)
    failing_api_mock.get.assert_called_once()



from django.urls import reverse
from rest_framework import serializers

from ..core.serializers import MutableFields
from ..core.utils import format_plaintext_for_html
from .models import Category

__all__ = ["CategorySerializer"]


def last_activity_detail(f):
    """util for serializing last activity details"""

    def decorator(self, obj):
        if not obj.last_thread_id:
            return None

        acl = self.get_acl(obj)
        tested_acls = (
            acl.get("can_see"),
            acl.get("can_browse"),
            acl.get("can_see_all_threads"),
        )

        if not all(tested_acls):
            return None

        return f(self, obj)

    return decorator


class CategorySerializer(serializers.ModelSerializer, MutableFields):
    parent = serializers.PrimaryKeyRelatedField(read_only=True)
    description = serializers.SerializerMethodField()
    is_read = serializers.SerializerMethodField()
    subcategories = serializers.SerializerMethodField()
    acl = serializers.SerializerMethodField()

    url = serializers.SerializerMethodField()

    class Meta:
        model = Category
        fields = [
            "id",
            "parent",
            "name",
            "short_name",
            "color",
            "description",
            "is_closed",
            "threads",
            "posts",
            "last_post_on",
            "last_thread_title",
            "last_poster",
            "last_poster_name",
            "css_class",
            "is_read",
            "subcategories",
            "acl",
            "level",
            "lft",
            "rght",
            "url",
        ]

    def to_representation(self, instance):
        data = super().to_representation(instance)
        if instance.special_role:
            data["special_role"] = instance.special_role

        return data

    def get_description(self, obj):
        if obj.description:
            return {
                "plain": obj.description,
                "html": format_plaintext_for_html(obj.description),
            }

        # Serve root category's description
        if obj.special_role == "root_category" and self.context:
            settings = self.context.settings
            if settings.index_message:
                return {
                    "plain": settings.index_message,
                    "html": format_plaintext_for_html(settings.index_message),
                }

    def get_css_class(self, obj):
        return obj.css_class or None

    def get_is_read(self, obj):
        try:
            return obj.is_read
        except AttributeError:
            return None

    def get_subcategories(self, obj):
        try:
            return CategorySerializer(
                obj.subcategories, context=self.context, many=True
            ).data
        except AttributeError:
            return []

    def get_acl(self, obj):
        try:
            return obj.acl
        except AttributeError:
            return {}

    @last_activity_detail
    def get_last_poster(self, obj):
        if obj.last_poster_id:
            return {
                "id": obj.last_poster_id,
                "avatars": obj.last_poster.avatars,
                "url": reverse(
                    "misago:user",
                    kwargs={"slug": obj.last_poster_slug, "pk": obj.last_poster_id},
                ),
            }

    def get_url(self, obj):
        return {
            "index": obj.get_absolute_url(),
            "last_thread": self.get_last_thread_url(obj),
            "last_thread_new": self.get_last_thread_new_url(obj),
            "last_post": self.get_last_post_url(obj),
        }

    @last_activity_detail
    def get_last_thread_url(self, obj):
        return obj.get_last_thread_url()

    @last_activity_detail
    def get_last_thread_new_url(self, obj):
        return obj.get_last_thread_new_url()

    @last_activity_detail
    def get_last_post_url(self, obj):
        return obj.get_last_post_url()


class CategoryWithPosterSerializer(CategorySerializer):
    last_poster = serializers.SerializerMethodField()

    def get_subcategories(self, obj):
        try:
            return CategoryWithPosterSerializer(obj.subcategories, many=True).data
        except AttributeError:
            return []


CategoryWithPosterSerializer = CategoryWithPosterSerializer.extend_fields("last_poster")



from django.db import migrations


class Migration(migrations.Migration):
    """Superseded by 0004"""

    dependencies = [("misago_acl", "0001_initial"), ("misago_core", "0001_initial")]

    operations = []



from unittest.mock import Mock

from ..context_processors import conf


def test_request_settings_are_included_in_template_context(db, dynamic_settings):
    mock_request = Mock(settings=dynamic_settings)
    context_settings = conf(mock_request)["settings"]
    assert context_settings == mock_request.settings


def test_settings_are_included_in_frontend_context(db, client):
    response = client.get("/")
    assert response.status_code == 200
    assert '"SETTINGS": {"' in response.content.decode("utf-8")



apipatterns = []
urlpatterns = []
context_processors = []

new_registrations_validators = []

oauth2_user_data_filters = []
oauth2_validators = []

post_search_filters = []
post_validators = []

markdown_extensions = []
parsing_result_processors = []



from ..pluginlist import parse_plugins_list


def test_empty_plugins_str_is_parsed_to_empty_list():
    assert parse_plugins_list("") == []


def test_comment_str_is_parsed_to_empty_list():
    assert parse_plugins_list("# comment") == []


def test_line_containing_plugin_name_is_parsed():
    assert parse_plugins_list("plugin") == ["plugin"]


def test_line_containing_local_plugin_name_is_parsed():
    assert parse_plugins_list("plugin@/local/") == ["plugin@/local/"]


def test_whitespace_is_stripped_from_local_plugin_name():
    assert parse_plugins_list("plugin @/local/") == ["plugin@/local/"]


def test_whitespace_is_stripped_from_local_plugin_path():
    assert parse_plugins_list("plugin@ /local/") == ["plugin@/local/"]


def test_comment_is_removed_from_line_containing_plugin_name():
    assert parse_plugins_list("plugin # comment") == ["plugin"]


def test_multiple_lines_containing_plugin_names_are_parsed():
    assert parse_plugins_list("plugin1\nplugin2") == ["plugin1", "plugin2"]


def test_empty_lines_are_skipped_by_parser():
    assert parse_plugins_list("plugin1\n\nplugin2") == ["plugin1", "plugin2"]


def test_comments_are_filtered_from_plugin_list():
    assert parse_plugins_list("plugin1\n# comment\nplugin2") == ["plugin1", "plugin2"]


def test_whitespace_is_stripped_from_line_start_and_end_by_parser():
    assert parse_plugins_list("plugin1\n  plugin2") == ["plugin1", "plugin2"]



from django.conf import settings
from django.core.checks import register, Critical

SUPPORTED_ENGINES = [
    "django.db.backends.postgresql",
    "django.db.backends.postgresql_psycopg2",
]


@register()
def check_db_engine(app_configs, **kwargs):
    errors = []

    try:
        if settings.DATABASES["default"]["ENGINE"] not in SUPPORTED_ENGINES:
            raise ValueError()
    except (AttributeError, KeyError, ValueError):
        errors.append(
            Critical(msg="Misago requires PostgreSQL database.", id="misago.001")
        )

    return errors



from django.shortcuts import render
from django.urls import reverse
from django.views import View

from ..viewmodels import ForumThread, PrivateThread, ThreadPosts


class ThreadBase(View):
    thread = None
    posts = ThreadPosts

    template_name = None

    def get(self, request, pk, slug, page=0):
        thread = self.get_thread(request, pk, slug)
        posts = self.get_posts(request, thread, page)

        frontend_context = self.get_frontend_context(request, thread, posts)
        request.frontend_context.update(frontend_context)

        template_context = self.get_template_context(request, thread, posts)
        return render(request, self.template_name, template_context)

    def get_thread(self, request, pk, slug):
        return self.thread(  # pylint: disable=not-callable
            request,
            pk,
            slug,
            path_aware=True,
            read_aware=True,
            subscription_aware=True,
            poll_votes_aware=True,
        )

    def get_posts(self, request, thread, page):
        return self.posts(request, thread, page)

    def get_default_frontend_context(self):
        return {}

    def get_frontend_context(self, request, thread, posts):
        context = self.get_default_frontend_context()

        context.update(
            {
                "THREAD": thread.get_frontend_context(),
                "POSTS": posts.get_frontend_context(),
            }
        )

        return context

    def get_template_context(self, request, thread, posts):
        context = {
            "url_name": ":".join(
                request.resolver_match.namespaces + [request.resolver_match.url_name]
            )
        }

        context.update(thread.get_template_context())
        context.update(posts.get_template_context())

        return context


class ThreadView(ThreadBase):
    thread = ForumThread
    template_name = "misago/thread/thread.html"

    def get_default_frontend_context(self):
        return {"THREADS_API": reverse("misago:api:thread-list")}


class PrivateThreadView(ThreadBase):
    thread = PrivateThread
    template_name = "misago/thread/private_thread.html"



import os
import random

PHRASES_FILE = os.path.join(os.path.dirname(os.path.abspath(__file__)), "phrases.txt")


class EnglishCorpus:
    def __init__(self, phrases_file=PHRASES_FILE, min_length=None, max_length=None):
        self._countdown = 0
        self._previous = None

        self.phrases = []
        with open(phrases_file, "r") as f:
            for phrase in [l.strip() for l in f.readlines()]:
                if min_length and len(phrase) < min_length:
                    continue
                if max_length and len(phrase) > max_length:
                    continue
                self.phrases.append(phrase)

    def _countdown_to_shuffle(self):
        self._countdown -= 1
        if self._countdown < 0:
            self._countdown = random.randint(500, 1000)
            self.shuffle()

    def __len__(self):
        return len(self.phrases)

    def shuffle(self):
        random.shuffle(self.phrases)

    def random_sentence(self):
        self._countdown_to_shuffle()

        choice = None
        while not choice or choice == self._previous:
            choice = random.choice(self.phrases)

        self._previous = choice
        return choice

    def random_sentences(self, no):
        self._countdown_to_shuffle()

        max_no = len(self) - no - 1
        start = random.randint(0, max_no)

        sentences = self.phrases[start : (start + no)]
        random.shuffle(sentences)

        return sentences



from django.core.exceptions import PermissionDenied
from django.utils.translation import gettext as _
from django.utils.translation import gettext_lazy
from rest_framework import serializers

from . import PostingEndpoint, PostingMiddleware
from ....acl.objectacl import add_acl_to_obj
from ....categories import THREADS_ROOT_NAME
from ....categories.models import Category
from ....categories.permissions import can_browse_category, can_see_category
from ...permissions import allow_start_thread
from ...threadtypes import trees_map


class CategoryMiddleware(PostingMiddleware):
    """
    middleware that validates category id and sets category on thread and post instances
    """

    def use_this_middleware(self):
        if self.mode == PostingEndpoint.START:
            return self.tree_name == THREADS_ROOT_NAME
        return False

    def get_serializer(self):
        return CategorySerializer(self.user_acl, data=self.request.data)

    def pre_save(self, serializer):
        category = serializer.category_cache

        add_acl_to_obj(self.user_acl, category)

        # set flags for savechanges middleware
        category.update_all = False
        category.update_fields = []

        # assign category to thread and post
        self.thread.category = category
        self.post.category = category


class CategorySerializer(serializers.Serializer):
    category = serializers.IntegerField(
        error_messages={
            "required": gettext_lazy("You have to select category to post thread in."),
            "invalid": gettext_lazy("Selected category is invalid."),
        }
    )

    def __init__(self, user_acl, *args, **kwargs):
        self.user_acl = user_acl
        self.category_cache = None

        super().__init__(*args, **kwargs)

    def validate_category(self, value):
        try:
            self.category_cache = Category.objects.get(
                pk=value, tree_id=trees_map.get_tree_id_for_root(THREADS_ROOT_NAME)
            )

            can_see = can_see_category(self.user_acl, self.category_cache)
            can_browse = can_browse_category(self.user_acl, self.category_cache)
            if not (self.category_cache.level and can_see and can_browse):
                raise PermissionDenied(_("Selected category is invalid."))

            allow_start_thread(self.user_acl, self.category_cache)
        except Category.DoesNotExist:
            raise serializers.ValidationError(
                _(
                    "Selected category doesn't exist or "
                    "you don't have permission to browse it."
                )
            )
        except PermissionDenied as e:
            raise serializers.ValidationError(e.args[0])



from django.http import Http404
from rest_framework.response import Response


def paginate(
    object_list,
    page,
    per_page,
    orphans=0,
    allow_empty_first_page=True,
    allow_explicit_first_page=False,
    paginator=None,
):
    from django.core.paginator import Paginator, EmptyPage, InvalidPage
    from .exceptions import ExplicitFirstPage

    if page in (1, "1") and not allow_explicit_first_page:
        raise ExplicitFirstPage()
    if not page:
        page = 1

    paginator = paginator or Paginator

    try:
        return paginator(
            object_list,
            per_page,
            orphans=orphans,
            allow_empty_first_page=allow_empty_first_page,
        ).page(page)
    except (EmptyPage, InvalidPage):
        raise Http404()


def pagination_dict(page):
    pagination = {
        "page": page.number,
        "pages": page.paginator.num_pages,
        "count": page.paginator.count,
        "first": None,
        "previous": None,
        "next": None,
        "last": None,
        "before": 0,
        "more": 0,
    }

    if page.has_previous():
        pagination["first"] = 1
        pagination["previous"] = page.previous_page_number()

    if page.has_next():
        pagination["last"] = page.paginator.num_pages
        pagination["next"] = page.next_page_number()

    if page.start_index():
        pagination["before"] = page.start_index() - 1
    pagination["more"] = page.paginator.count - page.end_index()

    return pagination


def paginated_response(page, serializer=None, data=None, extra=None):
    response_data = pagination_dict(page)

    results = list(data or page.object_list)
    if serializer:
        results = serializer(results, many=True).data

    response_data.update({"results": results})

    if extra:
        response_data.update(extra)

    return Response(response_data)


def validate_slug(model, slug):
    from .exceptions import OutdatedSlug

    if model.slug != slug:
        raise OutdatedSlug(model)


def get_int_or_404(value):
    if str(value).isdigit():
        return int(value)
    raise Http404()



from django.urls import reverse

from ....acl import ACL_CACHE
from ....admin.test import AdminTestCase
from ....cache.test import assert_invalidates_cache
from ....threads import test
from ....threads.models import Thread
from ...models import Category


class CategoryAdminTestCase(AdminTestCase):
    def assertValidTree(self, expected_tree):
        root = Category.objects.root_category()
        queryset = Category.objects.filter(tree_id=root.tree_id).order_by("lft")

        current_tree = []
        for category in queryset:
            current_tree.append(
                (
                    category,
                    category.level,
                    category.lft - root.lft + 1,
                    category.rght - root.lft + 1,
                )
            )

        if len(expected_tree) != len(current_tree):
            self.fail(
                "nodes tree is %s items long, should be %s"
                % (len(current_tree), len(expected_tree))
            )

        for i, category in enumerate(expected_tree):
            _category = current_tree[i]
            if category[0] != _category[0]:
                self.fail(
                    ("expected category at index #%s to be %s, " "found %s instead")
                    % (i, category[0], _category[0])
                )
            if category[1] != _category[1]:
                self.fail(
                    ("expected level at index #%s to be %s, " "found %s instead")
                    % (i, category[1], _category[1])
                )
            if category[2] != _category[2]:
                self.fail(
                    ("expected lft at index #%s to be %s, " "found %s instead")
                    % (i, category[2], _category[2])
                )
            if category[3] != _category[3]:
                self.fail(
                    ("expected lft at index #%s to be %s, " "found %s instead")
                    % (i, category[3], _category[3])
                )


class CategoryAdminViewsTests(CategoryAdminTestCase):
    def test_link_registered(self):
        """admin nav contains categories link"""
        response = self.client.get(reverse("misago:admin:categories:index"))

        self.assertContains(response, reverse("misago:admin:categories:index"))

    def test_list_view(self):
        """categories list view returns 200"""
        response = self.client.get(reverse("misago:admin:categories:index"))

        self.assertContains(response, "First category")

        # Now test that empty categories list contains message
        root = Category.objects.root_category()
        for descendant in root.get_descendants():
            descendant.delete()

        response = self.client.get(reverse("misago:admin:categories:index"))

        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "No categories")

    def test_new_view(self):
        """new category view has no showstoppers"""
        root = Category.objects.root_category()
        first_category = Category.objects.get(slug="first-category")

        response = self.client.get(reverse("misago:admin:categories:new"))
        self.assertEqual(response.status_code, 200)

        response = self.client.post(
            reverse("misago:admin:categories:new"),
            data={
                "name": "Test Category",
                "description": "Lorem ipsum dolor met",
                "new_parent": root.pk,
                "prune_started_after": 0,
                "prune_replied_after": 0,
            },
        )
        self.assertEqual(response.status_code, 302)

        response = self.client.get(reverse("misago:admin:categories:index"))
        self.assertContains(response, "Test Category")

        test_category = Category.objects.get(slug="test-category")

        self.assertValidTree(
            [(root, 0, 1, 6), (first_category, 1, 2, 3), (test_category, 1, 4, 5)]
        )

        response = self.client.post(
            reverse("misago:admin:categories:new"),
            data={
                "name": "Test Other Category",
                "description": "Lorem ipsum dolor met",
                "new_parent": root.pk,
                "prune_started_after": 0,
                "prune_replied_after": 0,
            },
        )
        self.assertEqual(response.status_code, 302)

        test_other_category = Category.objects.get(slug="test-other-category")

        self.assertValidTree(
            [
                (root, 0, 1, 8),
                (first_category, 1, 2, 3),
                (test_category, 1, 4, 5),
                (test_other_category, 1, 6, 7),
            ]
        )

        response = self.client.post(
            reverse("misago:admin:categories:new"),
            data={
                "name": "Test Subcategory",
                "new_parent": test_category.pk,
                "copy_permissions": test_category.pk,
                "prune_started_after": 0,
                "prune_replied_after": 0,
            },
        )
        self.assertEqual(response.status_code, 302)

        test_subcategory = Category.objects.get(slug="test-subcategory")

        self.assertValidTree(
            [
                (root, 0, 1, 10),
                (first_category, 1, 2, 3),
                (test_category, 1, 4, 7),
                (test_subcategory, 2, 5, 6),
                (test_other_category, 1, 8, 9),
            ]
        )

        response = self.client.get(reverse("misago:admin:categories:index"))
        self.assertContains(response, "Test Subcategory")

    def test_creating_new_category_invalidates_acl_cache(self):
        root = Category.objects.root_category()

        with assert_invalidates_cache(ACL_CACHE):
            self.client.post(
                reverse("misago:admin:categories:new"),
                data={
                    "name": "Test Category",
                    "description": "Lorem ipsum dolor met",
                    "new_parent": root.pk,
                    "prune_started_after": 0,
                    "prune_replied_after": 0,
                },
            )

    def test_edit_view(self):
        """edit category view has no showstoppers"""
        private_threads = Category.objects.private_threads()
        root = Category.objects.root_category()
        first_category = Category.objects.get(slug="first-category")

        response = self.client.get(
            reverse("misago:admin:categories:edit", kwargs={"pk": private_threads.pk})
        )
        self.assertEqual(response.status_code, 302)

        response = self.client.get(
            reverse("misago:admin:categories:edit", kwargs={"pk": root.pk})
        )
        self.assertEqual(response.status_code, 302)

        response = self.client.post(
            reverse("misago:admin:categories:new"),
            data={
                "name": "Test Category",
                "description": "Lorem ipsum dolor met",
                "new_parent": root.pk,
                "prune_started_after": 0,
                "prune_replied_after": 0,
            },
        )
        self.assertEqual(response.status_code, 302)

        test_category = Category.objects.get(slug="test-category")

        response = self.client.get(
            reverse("misago:admin:categories:edit", kwargs={"pk": test_category.pk})
        )

        self.assertContains(response, "Test Category")

        response = self.client.post(
            reverse("misago:admin:categories:edit", kwargs={"pk": test_category.pk}),
            data={
                "name": "Test Category Edited",
                "new_parent": root.pk,
                "role": "category",
                "prune_started_after": 0,
                "prune_replied_after": 0,
            },
        )
        self.assertEqual(response.status_code, 302)

        self.assertValidTree(
            [(root, 0, 1, 6), (first_category, 1, 2, 3), (test_category, 1, 4, 5)]
        )

        response = self.client.get(reverse("misago:admin:categories:index"))
        self.assertContains(response, "Test Category Edited")

        response = self.client.post(
            reverse("misago:admin:categories:edit", kwargs={"pk": test_category.pk}),
            data={
                "name": "Test Category Edited",
                "new_parent": first_category.pk,
                "role": "category",
                "prune_started_after": 0,
                "prune_replied_after": 0,
            },
        )
        self.assertEqual(response.status_code, 302)

        self.assertValidTree(
            [(root, 0, 1, 6), (first_category, 1, 2, 5), (test_category, 2, 3, 4)]
        )

        response = self.client.get(reverse("misago:admin:categories:index"))
        self.assertContains(response, "Test Category Edited")

    def test_editing_category_invalidates_acl_cache(self):
        root = Category.objects.root_category()
        self.client.post(
            reverse("misago:admin:categories:new"),
            data={
                "name": "Test Category",
                "description": "Lorem ipsum dolor met",
                "new_parent": root.pk,
                "prune_started_after": 0,
                "prune_replied_after": 0,
            },
        )

        test_category = Category.objects.get(slug="test-category")

        with assert_invalidates_cache(ACL_CACHE):
            self.client.post(
                reverse(
                    "misago:admin:categories:edit", kwargs={"pk": test_category.pk}
                ),
                data={
                    "name": "Test Category Edited",
                    "new_parent": root.pk,
                    "role": "category",
                    "prune_started_after": 0,
                    "prune_replied_after": 0,
                },
            )

    def test_move_views(self):
        """move up/down views have no showstoppers"""
        root = Category.objects.root_category()
        first_category = Category.objects.get(slug="first-category")

        self.client.post(
            reverse("misago:admin:categories:new"),
            data={
                "name": "Category A",
                "new_parent": root.pk,
                "prune_started_after": 0,
                "prune_replied_after": 0,
            },
        )
        category_a = Category.objects.get(slug="category-a")

        self.client.post(
            reverse("misago:admin:categories:new"),
            data={
                "name": "Category B",
                "new_parent": root.pk,
                "prune_started_after": 0,
                "prune_replied_after": 0,
            },
        )
        category_b = Category.objects.get(slug="category-b")

        response = self.client.post(
            reverse("misago:admin:categories:up", kwargs={"pk": category_b.pk})
        )
        self.assertEqual(response.status_code, 302)

        self.assertValidTree(
            [
                (root, 0, 1, 8),
                (first_category, 1, 2, 3),
                (category_b, 1, 4, 5),
                (category_a, 1, 6, 7),
            ]
        )

        response = self.client.post(
            reverse("misago:admin:categories:up", kwargs={"pk": category_b.pk})
        )
        self.assertEqual(response.status_code, 302)

        self.assertValidTree(
            [
                (root, 0, 1, 8),
                (category_b, 1, 2, 3),
                (first_category, 1, 4, 5),
                (category_a, 1, 6, 7),
            ]
        )

        response = self.client.post(
            reverse("misago:admin:categories:down", kwargs={"pk": category_b.pk})
        )
        self.assertEqual(response.status_code, 302)

        self.assertValidTree(
            [
                (root, 0, 1, 8),
                (first_category, 1, 2, 3),
                (category_b, 1, 4, 5),
                (category_a, 1, 6, 7),
            ]
        )

        response = self.client.post(
            reverse("misago:admin:categories:down", kwargs={"pk": category_b.pk})
        )
        self.assertEqual(response.status_code, 302)

        self.assertValidTree(
            [
                (root, 0, 1, 8),
                (first_category, 1, 2, 3),
                (category_a, 1, 4, 5),
                (category_b, 1, 6, 7),
            ]
        )

        response = self.client.post(
            reverse("misago:admin:categories:down", kwargs={"pk": category_b.pk})
        )
        self.assertEqual(response.status_code, 302)

        self.assertValidTree(
            [
                (root, 0, 1, 8),
                (first_category, 1, 2, 3),
                (category_a, 1, 4, 5),
                (category_b, 1, 6, 7),
            ]
        )


class CategoryAdminDeleteViewTests(CategoryAdminTestCase):
    def setUp(self):
        """
        Create categories tree for test cases:

        First category (created by migration)

        Category A
          + Category B
            + Subcategory C
            + Subcategory D

        Category E
          + Category F
        """

        super().setUp()

        self.root = Category.objects.root_category()
        self.first_category = Category.objects.get(slug="first-category")

        self.client.post(
            reverse("misago:admin:categories:new"),
            data={
                "name": "Category A",
                "new_parent": self.root.pk,
                "prune_started_after": 0,
                "prune_replied_after": 0,
            },
        )

        self.client.post(
            reverse("misago:admin:categories:new"),
            data={
                "name": "Category E",
                "new_parent": self.root.pk,
                "prune_started_after": 0,
                "prune_replied_after": 0,
            },
        )

        self.category_a = Category.objects.get(slug="category-a")
        self.category_e = Category.objects.get(slug="category-e")

        self.client.post(
            reverse("misago:admin:categories:new"),
            data={
                "name": "Category B",
                "new_parent": self.category_a.pk,
                "prune_started_after": 0,
                "prune_replied_after": 0,
            },
        )
        self.category_b = Category.objects.get(slug="category-b")

        self.client.post(
            reverse("misago:admin:categories:new"),
            data={
                "name": "Subcategory C",
                "new_parent": self.category_b.pk,
                "prune_started_after": 0,
                "prune_replied_after": 0,
            },
        )
        self.category_c = Category.objects.get(slug="subcategory-c")

        self.client.post(
            reverse("misago:admin:categories:new"),
            data={
                "name": "Subcategory D",
                "new_parent": self.category_b.pk,
                "prune_started_after": 0,
                "prune_replied_after": 0,
            },
        )
        self.category_d = Category.objects.get(slug="subcategory-d")

        self.client.post(
            reverse("misago:admin:categories:new"),
            data={
                "name": "Category F",
                "new_parent": self.category_e.pk,
                "prune_started_after": 0,
                "prune_replied_after": 0,
            },
        )
        self.category_f = Category.objects.get(slug="category-f")

    def test_delete_category_move_contents(self):
        """category was deleted and its contents were moved"""
        for _ in range(10):
            test.post_thread(self.category_b)
        self.assertEqual(Thread.objects.count(), 10)

        response = self.client.get(
            reverse("misago:admin:categories:delete", kwargs={"pk": self.category_b.pk})
        )
        self.assertEqual(response.status_code, 200)

        response = self.client.post(
            reverse(
                "misago:admin:categories:delete", kwargs={"pk": self.category_b.pk}
            ),
            data={
                "move_children_to": self.category_e.pk,
                "move_threads_to": self.category_d.pk,
            },
        )
        self.assertEqual(response.status_code, 302)
        self.assertEqual(Category.objects.all_categories().count(), 6)
        self.assertEqual(Thread.objects.count(), 10)
        for thread in Thread.objects.all():
            self.assertEqual(thread.category_id, self.category_d.pk)

        self.assertValidTree(
            [
                (self.root, 0, 1, 14),
                (self.first_category, 1, 2, 3),
                (self.category_a, 1, 4, 5),
                (self.category_e, 1, 6, 13),
                (self.category_f, 2, 7, 8),
                (self.category_c, 2, 9, 10),
                (self.category_d, 2, 11, 12),
            ]
        )

    def test_delete_category_and_contents(self):
        """category and its contents were deleted"""
        for _ in range(10):
            test.post_thread(self.category_b)

        response = self.client.get(
            reverse("misago:admin:categories:delete", kwargs={"pk": self.category_b.pk})
        )
        self.assertEqual(response.status_code, 200)

        response = self.client.post(
            reverse(
                "misago:admin:categories:delete", kwargs={"pk": self.category_b.pk}
            ),
            data={"move_children_to": "", "move_threads_to": ""},
        )
        self.assertEqual(response.status_code, 302)

        self.assertEqual(Category.objects.all_categories().count(), 4)
        self.assertEqual(Thread.objects.count(), 0)

        self.assertValidTree(
            [
                (self.root, 0, 1, 10),
                (self.first_category, 1, 2, 3),
                (self.category_a, 1, 4, 5),
                (self.category_e, 1, 6, 9),
                (self.category_f, 2, 7, 8),
            ]
        )

    def test_delete_leaf_category_and_contents(self):
        """leaf category was deleted with contents"""
        for _ in range(10):
            test.post_thread(self.category_d)
        self.assertEqual(Thread.objects.count(), 10)

        response = self.client.get(
            reverse("misago:admin:categories:delete", kwargs={"pk": self.category_d.pk})
        )
        self.assertEqual(response.status_code, 200)

        response = self.client.post(
            reverse(
                "misago:admin:categories:delete", kwargs={"pk": self.category_d.pk}
            ),
            data={"move_children_to": "", "move_threads_to": ""},
        )
        self.assertEqual(response.status_code, 302)

        self.assertEqual(Category.objects.all_categories().count(), 6)
        self.assertEqual(Thread.objects.count(), 0)

        self.assertValidTree(
            [
                (self.root, 0, 1, 14),
                (self.first_category, 1, 2, 3),
                (self.category_a, 1, 4, 9),
                (self.category_b, 2, 5, 8),
                (self.category_c, 3, 6, 7),
                (self.category_e, 1, 10, 13),
                (self.category_f, 2, 11, 12),
            ]
        )

    def test_deleting_category_invalidates_acl_cache(self):
        with assert_invalidates_cache(ACL_CACHE):
            self.client.post(
                reverse(
                    "misago:admin:categories:delete", kwargs={"pk": self.category_d.pk}
                ),
                data={"move_children_to": "", "move_threads_to": ""},
            )



from django.utils.translation import gettext as _
from rest_framework import status
from rest_framework.response import Response

from ....core.mail import mail_user
from ...credentialchange import store_new_credential
from ...serializers import ChangePasswordSerializer


def change_password_endpoint(request, pk=None):
    serializer = ChangePasswordSerializer(
        data=request.data, context={"user": request.user}
    )

    if not serializer.is_valid():
        return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)

    token = store_new_credential(
        request, "password", serializer.validated_data["new_password"]
    )

    mail_subject = _("Confirm password change on %(forum_name)s forums")
    mail_subject = mail_subject % {"forum_name": request.settings.forum_name}

    mail_user(
        request.user,
        mail_subject,
        "misago/emails/change_password",
        context={"settings": request.settings, "token": token},
    )

    return Response(
        {"detail": _("Password change confirmation link was sent to your address.")}
    )



from ..conf import settings

MOMENT_STATIC_PATH = "misago/momentjs/%s.js"


def get_locale_url(language, *, static_path_template=None, locales=None):
    locales = locales or settings.MISAGO_MOMENT_JS_LOCALES
    clean_language = clean_language_name(language, locales)
    if clean_language:
        static_path_template = static_path_template or MOMENT_STATIC_PATH
        return static_path_template % clean_language


def clean_language_name(language, locales):
    # lowercase language
    language = language.lower().replace("_", "-")

    # first try: literal match
    if language in locales:
        return language

    # second try: fallback to macrolanguage
    language = language.split("-")[0]
    if language in locales:
        return language

    # nothing was found
    return None



from ...models import Role
from ..test import mock_role_form_data


def test_factory_for_change_role_permissions_form_data():
    test_data = mock_role_form_data(Role(), {"can_fly": 1})
    assert "can_fly" in test_data



import pytest
from django.urls import reverse

from ....cache.test import assert_invalidates_cache
from ... import SOCIALAUTH_CACHE


@pytest.fixture
def admin_link(provider):
    return reverse(
        "misago:admin:settings:socialauth:disable", kwargs={"pk": provider.pk}
    )


def test_provider_can_be_disabled(admin_client, admin_link, provider):
    admin_client.post(admin_link)
    provider.refresh_from_db()
    assert not provider.is_active


def test_disabling_provider_invalidates_cache(admin_client, admin_link):
    with assert_invalidates_cache(SOCIALAUTH_CACHE):
        admin_client.post(admin_link)


def test_already_disabled_provider_cant_be_disabled(admin_client, disabled_provider):
    admin_link = reverse(
        "misago:admin:settings:socialauth:disable", kwargs={"pk": disabled_provider.pk}
    )
    admin_client.post(admin_link)
    disabled_provider.refresh_from_db()
    assert not disabled_provider.is_active


def test_provider_that_wasnt_setup_yet_cant_be_disabled(admin_client):
    admin_link = reverse(
        "misago:admin:settings:socialauth:disable", kwargs={"pk": "github"}
    )
    response = admin_client.post(admin_link)
    assert response.status_code == 302


def test_undefined_provider_cant_be_disabled(admin_client):
    admin_link = reverse(
        "misago:admin:settings:socialauth:disable", kwargs={"pk": "undefined"}
    )
    response = admin_client.post(admin_link)
    assert response.status_code == 302



import json

from django.urls import reverse

from ...users.test import AuthenticatedUserTestCase
from ..models import Agreement


class SubmitAgreementTests(AuthenticatedUserTestCase):
    def setUp(self):
        super().setUp()

        self.agreement = Agreement.objects.create(
            type=Agreement.TYPE_TOS, text="Lorem ipsum", is_active=True
        )

        self.api_link = reverse(
            "misago:api:submit-agreement", kwargs={"pk": self.agreement.pk}
        )

    def post_json(self, data):
        return self.client.post(
            self.api_link, json.dumps(data), content_type="application/json"
        )

    def test_anonymous(self):
        self.logout_user()

        response = self.client.post(self.api_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(), {"detail": "This action is not available to guests."}
        )

    def test_get_request(self):
        response = self.client.get(self.api_link)
        self.assertEqual(response.status_code, 405)
        self.assertEqual(response.json(), {"detail": 'Method "GET" not allowed.'})

    def test_invalid_agreement_id(self):
        api_link = reverse(
            "misago:api:submit-agreement", kwargs={"pk": self.agreement.pk + 1}
        )

        response = self.client.post(api_link)
        self.assertEqual(response.status_code, 404)
        self.assertEqual(response.json(), {"detail": "Not found."})

    def test_agreement_already_accepted(self):
        self.user.agreements.append(self.agreement.id)
        self.user.save()

        response = self.client.post(self.api_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(), {"detail": "You have already accepted this agreement."}
        )

    def test_no_accept_sent(self):
        response = self.client.post(self.api_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(), {"detail": "You need to submit a valid choice."}
        )

    def test_invalid_accept_sent(self):
        response = self.post_json({"accept": 1})
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(), {"detail": "You need to submit a valid choice."}
        )

    def test_accept_false(self):
        response = self.post_json({"accept": False})
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), {"detail": "ok"})

        self.user.refresh_from_db()
        self.assertTrue(self.user.is_deleting_account)
        self.assertFalse(self.user.is_active)

    def test_accept_false_staff(self):
        self.user.is_staff = True
        self.user.save()

        response = self.post_json({"accept": False})
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), {"detail": "ok"})

        self.user.refresh_from_db()
        self.assertFalse(self.user.is_deleting_account)
        self.assertTrue(self.user.is_active)

    def test_accept_true(self):
        response = self.post_json({"accept": True})
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), {"detail": "ok"})

        self.user.refresh_from_db()
        self.assertEqual(self.user.agreements, [self.agreement.id])
        self.assertFalse(self.user.is_deleting_account)
        self.assertTrue(self.user.is_active)



import copy

from . import buildacl
from .cache import get_acl_cache, set_acl_cache
from .providers import providers


def get_user_acl(user, cache_versions):
    user_acl = get_acl_cache(user, cache_versions)
    if user_acl is None:
        user_acl = buildacl.build_acl(user.get_roles())
        set_acl_cache(user, cache_versions, user_acl)
    user_acl["user_id"] = user.id
    user_acl["is_authenticated"] = bool(user.is_authenticated)
    user_acl["is_anonymous"] = bool(user.is_anonymous)
    user_acl["is_staff"] = user.is_staff
    user_acl["is_superuser"] = user.is_superuser
    user_acl["cache_versions"] = cache_versions.copy()
    return user_acl


def serialize_user_acl(user_acl):
    """serialize authenticated user's ACL"""
    serialized_acl = copy.deepcopy(user_acl)
    serialized_acl.pop("cache_versions")

    for serializer in providers.get_user_acl_serializers():
        serializer(serialized_acl)

    return serialized_acl



from django.contrib.auth import get_user_model
from django.utils.translation import gettext as _

from ..core.mail import build_mail, send_messages
from .events import record_event
from .models import ThreadParticipant

User = get_user_model()


def has_participants(thread):
    return thread.threadparticipant_set.exists()


def make_participants_aware(user, target):
    if hasattr(target, "__iter__"):
        make_threads_participants_aware(user, target)
    else:
        make_thread_participants_aware(user, target)


def make_threads_participants_aware(user, threads):
    threads_dict = {}
    for thread in threads:
        thread.participant = None
        threads_dict[thread.pk] = thread

    participants_qs = ThreadParticipant.objects.filter(
        user=user, thread_id__in=threads_dict.keys()
    )

    for participant in participants_qs:
        participant.user = user
        threads_dict[participant.thread_id].participant = participant


def make_thread_participants_aware(user, thread):
    thread.participants_list = []
    thread.participant = None

    participants_qs = ThreadParticipant.objects.filter(thread=thread)
    participants_qs = participants_qs.select_related("user")
    for participant in participants_qs.order_by("-is_owner", "user__slug"):
        participant.thread = thread
        thread.participants_list.append(participant)
        if participant.user == user:
            thread.participant = participant
    return thread.participants_list


def set_users_unread_private_threads_sync(
    users=None, participants=None, exclude_user=None
):
    users_ids = []
    if users:
        users_ids += [u.pk for u in users]
    if participants:
        users_ids += [p.user_id for p in participants]
    if exclude_user:
        users_ids = filter(lambda u: u != exclude_user.pk, users_ids)

    if not users_ids:
        return

    User.objects.filter(id__in=set(users_ids)).update(sync_unread_private_threads=True)


def set_owner(thread, user):
    ThreadParticipant.objects.set_owner(thread, user)


def change_owner(request, thread, new_owner):
    ThreadParticipant.objects.set_owner(thread, new_owner)
    set_users_unread_private_threads_sync(
        participants=thread.participants_list, exclude_user=request.user
    )

    if thread.participant and thread.participant.is_owner:
        record_event(
            request,
            thread,
            "changed_owner",
            {
                "user": {
                    "id": new_owner.id,
                    "username": new_owner.username,
                    "url": new_owner.get_absolute_url(),
                }
            },
        )
    else:
        record_event(request, thread, "tookover")


def add_participant(request, thread, new_participant):
    """adds single participant to thread, registers this on the event"""
    add_participants(request, thread, [new_participant])

    if request.user == new_participant:
        record_event(request, thread, "entered_thread")
    else:
        record_event(
            request,
            thread,
            "added_participant",
            {
                "user": {
                    "id": new_participant.id,
                    "username": new_participant.username,
                    "url": new_participant.get_absolute_url(),
                }
            },
        )


def add_participants(request, thread, users):
    """
    Add multiple participants to thread, set "recound private threads" flag on them
    notify them about being added to thread.
    """
    ThreadParticipant.objects.add_participants(thread, users)

    try:
        thread_participants = thread.participants_list
    except AttributeError:
        thread_participants = []

    set_users_unread_private_threads_sync(
        users=users, participants=thread_participants, exclude_user=request.user
    )

    emails = []
    for user in users:
        if user != request.user:
            emails.append(build_noticiation_email(request, thread, user))
    if emails:
        send_messages(emails)


def build_noticiation_email(request, thread, user):
    subject = _(
        '%(user)s has invited you to participate in private thread "%(thread)s"'
    )
    subject_formats = {"thread": thread.title, "user": request.user.username}

    return build_mail(
        user,
        subject % subject_formats,
        "misago/emails/privatethread/added",
        sender=request.user,
        context={"settings": request.settings, "thread": thread},
    )


def remove_participant(request, thread, user):
    """remove thread participant, set "recound private threads" flag on user"""
    removed_owner = False
    remaining_participants = []

    for participant in thread.participants_list:
        if participant.user == user:
            removed_owner = participant.is_owner
        else:
            remaining_participants.append(participant.user)

    set_users_unread_private_threads_sync(participants=thread.participants_list)

    if not remaining_participants:
        thread.delete()
    else:
        thread.threadparticipant_set.filter(user=user).delete()
        thread.subscription_set.filter(user=user).delete()

        if removed_owner:
            thread.is_closed = True  # flag thread to close

            if request.user == user:
                event_type = "owner_left"
            else:
                event_type = "removed_owner"
        else:
            if request.user == user:
                event_type = "participant_left"
            else:
                event_type = "removed_participant"

        record_event(
            request,
            thread,
            event_type,
            {
                "user": {
                    "id": user.id,
                    "username": user.username,
                    "url": user.get_absolute_url(),
                }
            },
        )



from django.utils.translation import gettext as _

from ..core.mail import mail_user
from ..legal.models import Agreement
from ..legal.utils import save_user_agreement_acceptance
from .tokens import make_activation_token


def send_welcome_email(request, user):
    settings = request.settings

    mail_subject = _("Welcome on %(forum_name)s forums!")
    mail_subject = mail_subject % {"forum_name": settings.forum_name}

    if not user.requires_activation:
        mail_user(
            user,
            mail_subject,
            "misago/emails/register/complete",
            context={"settings": settings},
        )
        return

    activation_token = make_activation_token(user)

    activation_by_admin = user.requires_activation_by_admin
    activation_by_user = user.requires_activation_by_user

    mail_user(
        user,
        mail_subject,
        "misago/emails/register/inactive",
        context={
            "activation_token": activation_token,
            "activation_by_admin": activation_by_admin,
            "activation_by_user": activation_by_user,
            "settings": settings,
        },
    )


def save_user_agreements(user, form):
    if not form.agreements:
        return

    for field_name in form.agreements.keys():
        agreement_id = form.cleaned_data[field_name]
        agreement = Agreement.objects.get(id=agreement_id)
        save_user_agreement_acceptance(user, agreement)

    user.save(update_fields=["agreements"])


def get_registration_result_json(user):
    activation_method = "active"
    if user.requires_activation_by_admin:
        activation_method = "admin"
    elif user.requires_activation_by_user:
        activation_method = "user"

    return {
        "activation": activation_method,
        "email": user.email,
        "username": user.username,
    }



# pylint: disable=super-init-not-called
from django.core.exceptions import PermissionDenied
from social_core.exceptions import AuthException


class AjaxError(Exception):
    """You've tried to do something over AJAX but misago blurped"""

    def __init__(self, message=None, code=406):
        self.message = message
        self.code = code


class Banned(PermissionDenied):
    def __init__(self, ban=None):
        if ban:
            self.ban = ban


class SocialAuthFailed(AuthException):
    """Exception used to return error messages from Misago's social auth to user."""

    def __init__(self, backend, message):
        self.backend = backend
        self.message = message


class SocialAuthBanned(AuthException):
    """Exception used to return ban message from Misago's social auth to user."""

    def __init__(self, backend, ban):
        self.backend = backend
        self.ban = ban


class ExplicitFirstPage(Exception):
    """The url that was used to reach view contained explicit first page"""


class OutdatedSlug(Exception):
    """The url that was used to reach view contained outdated slug"""



# Generated by Django 1.11.15 on 2018-08-15 20:58
import django.db.models.deletion
import django.utils.timezone
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("misago_legal", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="Agreement",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "type",
                    models.CharField(
                        choices=[
                            ("terms_of_service", "Terms of service"),
                            ("privacy_policy", "Privacy policy"),
                        ],
                        db_index=True,
                        default="terms_of_service",
                        max_length=20,
                    ),
                ),
                ("title", models.CharField(blank=True, max_length=255, null=True)),
                ("link", models.URLField(blank=True, max_length=255, null=True)),
                ("text", models.TextField(blank=True, null=True)),
                ("is_active", models.BooleanField(default=False)),
                ("created_on", models.DateTimeField(default=django.utils.timezone.now)),
                (
                    "created_by_name",
                    models.CharField(blank=True, max_length=255, null=True),
                ),
                ("last_modified_on", models.DateTimeField(blank=True, null=True)),
                (
                    "last_modified_by_name",
                    models.CharField(blank=True, max_length=255, null=True),
                ),
                (
                    "created_by",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "last_modified_by",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="UserAgreement",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "accepted_on",
                    models.DateTimeField(default=django.utils.timezone.now),
                ),
                (
                    "agreement",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="accepted_by",
                        to="misago_legal.Agreement",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={"ordering": ["-pk"]},
        ),
    ]



import time

from django.core.management.base import BaseCommand

from ....core.management.progressbar import show_progress
from ....core.pgutils import chunk_queryset
from ...models import Post


class Command(BaseCommand):
    help = "Rebuilds posts search"

    def handle(self, *args, **options):
        posts_to_reindex = Post.objects.filter(is_event=False).count()

        if not posts_to_reindex:
            self.stdout.write("\n\nNo posts were found")
        else:
            self.rebuild_posts_search(posts_to_reindex)

    def rebuild_posts_search(self, posts_to_reindex):
        self.stdout.write("Rebuilding search for %s posts...\n" % posts_to_reindex)

        rebuild_count = 0
        show_progress(self, rebuild_count, posts_to_reindex)
        start_time = time.time()

        queryset = Post.objects.select_related("thread").filter(is_event=False)
        for post in chunk_queryset(queryset):
            if post.id == post.thread.first_post_id:
                post.set_search_document(post.thread.title)
            else:
                post.set_search_document()
            post.save(update_fields=["search_document"])

            post.update_search_vector()
            post.save(update_fields=["search_vector"])

            rebuild_count += 1
            show_progress(self, rebuild_count, posts_to_reindex, start_time)

        self.stdout.write("\n\nRebuild search for %s posts" % rebuild_count)
